"""Azure resource-specific handlers for Terraform graph processing.

Handles special cases for Azure resources including resource groups, virtual networks,
network security groups, load balancers, and other Azure-specific relationships.
"""

from typing import Dict, List, Any
import modules.config.cloud_config_azure as cloud_config
import modules.helpers as helpers
from ast import literal_eval
import re
import copy

REVERSE_ARROW_LIST = cloud_config.AZURE_REVERSE_ARROW_LIST
IMPLIED_CONNECTIONS = cloud_config.AZURE_IMPLIED_CONNECTIONS
GROUP_NODES = cloud_config.AZURE_GROUP_NODES
CONSOLIDATED_NODES = cloud_config.AZURE_CONSOLIDATED_NODES
NODE_VARIANTS = cloud_config.AZURE_NODE_VARIANTS
SPECIAL_RESOURCES = cloud_config.AZURE_SPECIAL_RESOURCES
SHARED_SERVICES = cloud_config.AZURE_SHARED_SERVICES
DISCONNECT_SERVICES = cloud_config.AZURE_DISCONNECT_LIST


def handle_special_cases(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle special resource cases and disconnections for Azure.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with special cases handled
    """
    # Remove connections to services specified in disconnect services
    for r in sorted(tfdata["graphdict"].keys()):
        for d in DISCONNECT_SERVICES:
            if d in r:
                tfdata["graphdict"][r] = []
    return tfdata


def azure_handle_resource_group(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure Resource Group relationships.

    Resource Groups are top-level containers in Azure. All Azure resources
    belong to a Resource Group.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Resource Group relationships configured
    """
    # Find all resource groups
    resource_groups = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_resource_group"
    )

    if not resource_groups:
        return tfdata

    # For each resource, check if it has a resource_group_name attribute
    # and link it to the appropriate resource group
    for rg in resource_groups:
        rg_name = helpers.get_no_module_name(rg).split(".")[-1]

        # Find resources that reference this resource group
        for resource in sorted(tfdata["graphdict"].keys()):
            if resource == rg:
                continue

            resource_type = helpers.get_no_module_name(resource).split(".")[0]

            # Skip if already a child of this RG
            if resource in tfdata["graphdict"].get(rg, []):
                continue

            # Skip group nodes that are not VNets (VNets go directly under RG)
            if (
                resource_type in GROUP_NODES
                and resource_type != "azurerm_virtual_network"
            ):
                continue

            # Check metadata for resource_group_name reference
            if tfdata["meta_data"].get(resource):
                rg_ref = tfdata["meta_data"][resource].get("resource_group_name", "")
                # Check if this resource references our RG
                if rg in str(rg_ref) or rg_name in str(rg_ref):
                    # Add VNets directly under RG
                    if resource_type == "azurerm_virtual_network":
                        if resource not in tfdata["graphdict"][rg]:
                            tfdata["graphdict"][rg].append(resource)
                    # For non-VNet, non-group resources that aren't in a subnet yet
                    elif resource_type not in GROUP_NODES:
                        # Check if resource is already placed in a subnet or VNet
                        parent_list = helpers.list_of_parents(
                            tfdata["graphdict"], resource
                        )
                        in_hierarchy = any(
                            helpers.get_no_module_name(p).split(".")[0] in GROUP_NODES
                            for p in parent_list
                        )
                        if not in_hierarchy:
                            if resource not in tfdata["graphdict"][rg]:
                                tfdata["graphdict"][rg].append(resource)

    return tfdata


def azure_handle_vnet(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure Virtual Network relationships.

    VNets are network boundary containers within Resource Groups.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with VNet relationships configured
    """
    # Find all VNets
    vnets = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_virtual_network"
    )

    if not vnets:
        return tfdata

    # Find all subnets and link them to their VNets
    subnets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "azurerm_subnet")

    for vnet in vnets:
        vnet_name = helpers.get_no_module_name(vnet).split(".")[-1]

        for subnet in subnets:
            # Skip subnet associations
            if "association" in subnet:
                continue

            # Check if subnet references this VNet
            if tfdata["meta_data"].get(subnet):
                vnet_ref = tfdata["meta_data"][subnet].get("virtual_network_name", "")
                if vnet in str(vnet_ref) or vnet_name in str(vnet_ref):
                    if subnet not in tfdata["graphdict"].get(vnet, []):
                        tfdata["graphdict"][vnet].append(subnet)
                    # Remove subnet from other parents that aren't VNets
                    parent_list = helpers.list_of_parents(tfdata["graphdict"], subnet)
                    for parent in parent_list:
                        parent_type = helpers.get_no_module_name(parent).split(".")[0]
                        if parent != vnet and parent_type != "azurerm_virtual_network":
                            if subnet in tfdata["graphdict"].get(parent, []):
                                tfdata["graphdict"][parent].remove(subnet)

    return tfdata


def azure_handle_subnet(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure Subnet relationships.

    Subnets are network segments within VNets.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Subnet relationships configured
    """
    # Find all subnets (excluding association resources)
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]

    if not subnets:
        return tfdata

    # Find network interfaces and their subnet associations
    nics = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_network_interface"
    )

    for subnet in subnets:
        subnet_name = helpers.get_no_module_name(subnet).split(".")[-1]

        # Link NICs to subnets based on ip_configuration.subnet_id
        for nic in nics:
            if tfdata["meta_data"].get(nic):
                ip_config = tfdata["meta_data"][nic].get("ip_configuration", "")
                if subnet in str(ip_config) or subnet_name in str(ip_config):
                    if nic not in tfdata["graphdict"].get(subnet, []):
                        tfdata["graphdict"][subnet].append(nic)

        # NOTE: VM placement into subnets now happens in match_resources()
        # after create_multiple_resources() completes numbering

    return tfdata


def azure_handle_nsg(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure Network Security Group relationships.

    NSGs can be associated with Subnets or Network Interfaces.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with NSG relationships configured
    """
    # Find all NSGs
    nsgs = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_network_security_group"
    )

    # Filter out association resources from the NSG list
    nsgs = [n for n in nsgs if "association" not in n]

    if not nsgs:
        return tfdata

    # Find subnet-NSG associations
    subnet_nsg_associations = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_subnet_network_security_group_association"
    )

    # Find NIC-NSG associations
    nic_nsg_associations = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_network_interface_security_group_association"
    )

    # Process subnet-NSG associations
    for assoc in subnet_nsg_associations:
        if not tfdata["meta_data"].get(assoc):
            continue

        subnet_id = tfdata["meta_data"][assoc].get("subnet_id", "")
        nsg_id = tfdata["meta_data"][assoc].get("network_security_group_id", "")

        # Find the subnet and NSG
        target_subnet = None
        target_nsg = None

        for subnet in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        ):
            if "association" in subnet:
                continue
            if subnet in str(subnet_id) or subnet.split(".")[-1] in str(subnet_id):
                target_subnet = subnet
                break

        for nsg in nsgs:
            if nsg in str(nsg_id) or nsg.split(".")[-1] in str(nsg_id):
                target_nsg = nsg
                break

        # Link NSG to subnet - NSG becomes a container for subnet's resources
        if target_subnet and target_nsg:
            # Get resources currently in the subnet
            subnet_resources = list(tfdata["graphdict"].get(target_subnet, []))

            # Add NSG to subnet
            if target_nsg not in tfdata["graphdict"].get(target_subnet, []):
                tfdata["graphdict"][target_subnet].append(target_nsg)

            # Move non-group resources from subnet to NSG
            for resource in subnet_resources:
                resource_type = helpers.get_no_module_name(resource).split(".")[0]
                if resource_type not in GROUP_NODES and resource != target_nsg:
                    # Add resource to NSG
                    if resource not in tfdata["graphdict"].get(target_nsg, []):
                        tfdata["graphdict"][target_nsg].append(resource)
                    # Remove from direct subnet connection
                    if resource in tfdata["graphdict"].get(target_subnet, []):
                        tfdata["graphdict"][target_subnet].remove(resource)

        # Remove the association resource from graph (it's a linking resource)
        if assoc in tfdata["graphdict"]:
            del tfdata["graphdict"][assoc]

    # Process NIC-NSG associations
    for assoc in nic_nsg_associations:
        if not tfdata["meta_data"].get(assoc):
            continue

        nic_id = tfdata["meta_data"][assoc].get("network_interface_id", "")
        nsg_id = tfdata["meta_data"][assoc].get("network_security_group_id", "")

        # Find the NIC and NSG
        target_nic = None
        target_nsg = None

        for nic in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_network_interface"
        ):
            if nic in str(nic_id) or nic.split(".")[-1] in str(nic_id):
                target_nic = nic
                break

        for nsg in nsgs:
            if nsg in str(nsg_id) or nsg.split(".")[-1] in str(nsg_id):
                target_nsg = nsg
                break

        # Link NIC to NSG
        if target_nic and target_nsg:
            if target_nic not in tfdata["graphdict"].get(target_nsg, []):
                tfdata["graphdict"][target_nsg].append(target_nic)

        # Remove the association resource from graph
        if assoc in tfdata["graphdict"]:
            del tfdata["graphdict"][assoc]

    return tfdata


def azure_handle_vmss(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure VM Scale Set relationships and zone placement.

    Creates availability zone containers for VMSS instances that are expanded
    across multiple zones.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with VMSS relationships and zone containers configured
    """
    # Find all VMSS resources (including numbered instances)
    vmss_list = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_virtual_machine_scale_set"
    )

    if not vmss_list:
        return tfdata

    # Find subnets and load balancers
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]
    load_balancers = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_lb"
    )

    # Separate numbered and non-numbered VMSS
    numbered_vmss = [v for v in vmss_list if "~" in v]
    unnumbered_vmss = [v for v in vmss_list if "~" not in v]

    # Handle non-numbered VMSS (original logic)
    for vmss in unnumbered_vmss:
        if not tfdata["meta_data"].get(vmss):
            continue

        # Check network_profile for subnet references
        network_profile = tfdata["meta_data"][vmss].get("network_profile", "")

        # Link VMSS to subnet
        for subnet in subnets:
            subnet_name = subnet.split(".")[-1]
            if subnet in str(network_profile) or subnet_name in str(network_profile):
                if vmss not in tfdata["graphdict"].get(subnet, []):
                    tfdata["graphdict"][subnet].append(vmss)
                break

        # Check for load balancer backend pool associations
        lb_backend = tfdata["meta_data"][vmss].get(
            "load_balancer_backend_address_pool_ids", ""
        )
        for lb in load_balancers:
            lb_name = lb.split(".")[-1]
            if lb in str(lb_backend) or lb_name in str(lb_backend):
                # Link LB to VMSS
                if vmss not in tfdata["graphdict"].get(lb, []):
                    tfdata["graphdict"][lb].append(vmss)

    # Handle numbered VMSS - create zone containers
    if numbered_vmss:
        # Group numbered VMSS by base name
        vmss_by_base = {}
        for vmss in numbered_vmss:
            base_name = vmss.split("~")[0]
            if base_name not in vmss_by_base:
                vmss_by_base[base_name] = []
            vmss_by_base[base_name].append(vmss)

        for base_name, instances in vmss_by_base.items():
            # Get metadata from first instance
            if instances and tfdata["meta_data"].get(instances[0]):
                metadata = tfdata["meta_data"][instances[0]]
                zones_attr = metadata.get("zones", [])

                # Only create zone containers if zones attribute exists and has values
                if zones_attr and isinstance(zones_attr, list):
                    # Find parent subnet for this VMSS
                    parent_subnet = None
                    network_profile = metadata.get("network_profile", "")
                    for subnet in subnets:
                        subnet_name = subnet.split(".")[-1]
                        if subnet in str(network_profile) or subnet_name in str(
                            network_profile
                        ):
                            parent_subnet = subnet
                            break

                    if parent_subnet:
                        # Create zone container for each instance
                        for i, vmss_instance in enumerate(sorted(instances), start=1):
                            if i <= len(zones_attr):
                                zone_id = zones_attr[i - 1]
                                # Create zone node name
                                zone_node = f"tv_azurerm_zone.zone_{zone_id}"

                                # Create zone node in graphdict if it doesn't exist
                                if zone_node not in tfdata["graphdict"]:
                                    tfdata["graphdict"][zone_node] = []
                                    tfdata["meta_data"][zone_node] = {
                                        "zone_id": zone_id,
                                        "name": f"Availability Zone {zone_id}",
                                    }

                                # Place VMSS instance inside zone
                                if vmss_instance not in tfdata["graphdict"][zone_node]:
                                    tfdata["graphdict"][zone_node].append(vmss_instance)

                                # Remove VMSS instance from subnet direct children
                                if vmss_instance in tfdata["graphdict"].get(
                                    parent_subnet, []
                                ):
                                    tfdata["graphdict"][parent_subnet].remove(
                                        vmss_instance
                                    )

                                # Add zone to subnet if not already there
                                if zone_node not in tfdata["graphdict"].get(
                                    parent_subnet, []
                                ):
                                    tfdata["graphdict"][parent_subnet].append(zone_node)

    return tfdata


def azure_handle_appgw(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle Azure Application Gateway relationships.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Application Gateway relationships configured
    """
    # Find all Application Gateways
    appgws = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_application_gateway"
    )

    if not appgws:
        return tfdata

    # Find subnets
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]

    for appgw in appgws:
        if not tfdata["meta_data"].get(appgw):
            continue

        # Check gateway_ip_configuration for subnet references
        gateway_ip_config = tfdata["meta_data"][appgw].get(
            "gateway_ip_configuration", ""
        )

        # Link App Gateway to subnet
        for subnet in subnets:
            subnet_name = subnet.split(".")[-1]
            if subnet in str(gateway_ip_config) or subnet_name in str(
                gateway_ip_config
            ):
                # App Gateway typically goes in its own subnet, add it under subnet
                if appgw not in tfdata["graphdict"].get(subnet, []):
                    tfdata["graphdict"][subnet].append(appgw)
                break

        # Check backend_address_pool for backend references
        backend_pools = tfdata["meta_data"][appgw].get("backend_address_pool", "")

        # Find VMs, NICs, or IP addresses referenced in backend pools
        if backend_pools:
            # Look for FQDN or IP references to link to backend resources
            vms = helpers.list_of_dictkeys_containing(
                tfdata["graphdict"], "azurerm_virtual_machine"
            )
            for vm in vms:
                vm_name = vm.split(".")[-1]
                if vm in str(backend_pools) or vm_name in str(backend_pools):
                    if vm not in tfdata["graphdict"].get(appgw, []):
                        tfdata["graphdict"][appgw].append(vm)

            # Link to VMSS if referenced
            vmss_list = helpers.list_of_dictkeys_containing(
                tfdata["graphdict"], "azurerm_virtual_machine_scale_set"
            )
            for vmss in vmss_list:
                vmss_name = vmss.split(".")[-1]
                if vmss in str(backend_pools) or vmss_name in str(backend_pools):
                    if vmss not in tfdata["graphdict"].get(appgw, []):
                        tfdata["graphdict"][appgw].append(vmss)

    return tfdata


# azure_handle_sharedgroup REMOVED - Decision 9: Shared services grouping
# is now core functionality in graphmaker.py (group_shared_services_core)
# instead of a handler. This eliminates per-provider duplication.


def random_string_handler(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Remove random string resources from graph.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with random strings removed
    """
    randoms = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "random_string.")
    for r in list(randoms):
        del tfdata["graphdict"][r]
    return tfdata


def match_resources(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Match Azure resources based on patterns and dependencies.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with resources matched
    """
    # Match NSGs to Subnets by suffix pattern
    tfdata["graphdict"] = match_nsg_to_subnets(tfdata["graphdict"])

    # Match NICs to VMs
    tfdata["graphdict"] = match_nic_to_vm(
        tfdata["graphdict"], tfdata.get("meta_data", {})
    )

    # Place VMs into subnets (runs after numbering to handle numbered VMs correctly)
    tfdata = place_vms_in_subnets(tfdata)

    # Create availability zone containers for numbered VM instances
    # This must run AFTER place_vms_in_subnets so VMs are in subnet
    tfdata = create_vm_zone_containers(tfdata)

    # Create availability zone containers for numbered VMSS instances
    # This must run AFTER create_multiple_resources so numbered instances exist
    tfdata = create_vmss_zone_containers(tfdata)

    # Connect Load Balancers and Application Gateways to backend VMs
    tfdata = connect_lb_to_backend_vms(tfdata)

    # Clean up orphaned resources
    tfdata["graphdict"] = remove_empty_groups(tfdata["graphdict"])

    return tfdata


def create_vm_zone_containers(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Create availability zone containers for numbered VM instances.

    This function runs AFTER create_multiple_resources, so numbered VM instances
    like vm~1, vm~2, vm~3 already exist. It creates zone containers and places
    each instance in its corresponding zone.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with zone containers created
    """
    # Find all numbered VM instances (individual VMs, not VMSS)
    numbered_vms = [
        k
        for k in tfdata["graphdict"].keys()
        if (
            "azurerm_linux_virtual_machine" in k
            or "azurerm_windows_virtual_machine" in k
            or "azurerm_virtual_machine" in k
        )
        and "~" in k
        and "scale_set" not in k  # Exclude VMSS
    ]

    if not numbered_vms:
        return tfdata

    # Find subnets
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]

    # Group VMs by subnet
    vms_by_subnet = {}
    for vm in numbered_vms:
        # Find which subnet contains this VM
        vm_subnet = None
        for subnet in subnets:
            if vm in tfdata["graphdict"].get(subnet, []):
                vm_subnet = subnet
                break

        if vm_subnet:
            if vm_subnet not in vms_by_subnet:
                vms_by_subnet[vm_subnet] = []
            vms_by_subnet[vm_subnet].append(vm)

    # For each subnet with zoned VMs, create zone containers
    for subnet, vms in vms_by_subnet.items():
        # Check if VMs have zone attributes
        zones_used = {}
        for vm in vms:
            vm_metadata = tfdata["meta_data"].get(vm, {})
            zone = vm_metadata.get("zone", "")
            if zone:
                # Remove quotes if present
                zone = str(zone).strip('"').strip("'")

                # Resolve count.index expressions for numbered instances
                # e.g., "${tostring(count.index + 1)}" with vm~2 should become "2"
                if "count.index" in zone and "~" in vm:
                    # Extract instance number from vm~X
                    instance_num = int(vm.split("~")[1])
                    # count.index is 0-based, so instance 1 has count.index = 0
                    count_index = instance_num - 1
                    # Replace count.index with the actual value
                    zone = zone.replace("count.index", str(count_index))
                    # Evaluate simple expressions like "0 + 1" -> "1"
                    # Remove ${} and tostring() wrappers
                    zone = zone.replace("${", "").replace("}", "")
                    zone = zone.replace("tostring(", "").replace(")", "")
                    # Evaluate the expression
                    try:
                        zone = str(eval(zone))
                    except Exception:
                        # If evaluation fails, keep original
                        pass

                if zone not in zones_used:
                    zones_used[zone] = []
                zones_used[zone].append(vm)

        # If no zones found, skip this subnet
        if not zones_used:
            continue

        # Create zone containers
        for zone_id, zone_vms in zones_used.items():
            zone_name = f"tv_azurerm_zone.zone{zone_id}"

            # Create zone container node
            if zone_name not in tfdata["graphdict"]:
                tfdata["graphdict"][zone_name] = []
                tfdata["meta_data"][zone_name] = {
                    "zone_id": zone_id,
                    "type": "tv_azurerm_zone",
                }

            # Move VMs from subnet to zone
            for vm in zone_vms:
                if vm in tfdata["graphdict"].get(subnet, []):
                    tfdata["graphdict"][subnet].remove(vm)

                if vm not in tfdata["graphdict"][zone_name]:
                    tfdata["graphdict"][zone_name].append(vm)

            # Add zone to subnet
            if zone_name not in tfdata["graphdict"].get(subnet, []):
                tfdata["graphdict"][subnet].append(zone_name)

    return tfdata


def create_vmss_zone_containers(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Create availability zone containers for numbered VMSS instances.

    This function runs AFTER create_multiple_resources, so numbered instances
    like vmss~1, vmss~2, vmss~3 already exist. It creates zone containers and
    places each instance in its corresponding zone.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with zone containers created
    """
    # Find all numbered VMSS instances
    numbered_vmss = [
        k
        for k in tfdata["graphdict"].keys()
        if (
            "azurerm_linux_virtual_machine_scale_set" in k
            or "azurerm_windows_virtual_machine_scale_set" in k
            or "azurerm_virtual_machine_scale_set" in k
        )
        and "~" in k
    ]

    if not numbered_vmss:
        return tfdata

    # Find subnets
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]

    # Group numbered VMSS by base name
    vmss_by_base = {}
    for vmss in numbered_vmss:
        base_name = vmss.split("~")[0]
        if base_name not in vmss_by_base:
            vmss_by_base[base_name] = []
        vmss_by_base[base_name].append(vmss)

    for base_name, instances in vmss_by_base.items():
        # Get zones from metadata
        metadata = tfdata["meta_data"].get(base_name) or tfdata.get(
            "original_metadata", {}
        ).get(base_name, {})
        if not metadata:
            metadata = tfdata["meta_data"].get(instances[0], {})

        zones_attr_raw = metadata.get("zones", []) if metadata else []

        # zones might be a string representation of a list, convert it
        if isinstance(zones_attr_raw, str):
            try:
                zones_attr = literal_eval(zones_attr_raw)
            except:
                zones_attr = []
        else:
            zones_attr = zones_attr_raw

        # Get subnet reference from all_resource (original Terraform config)
        subnet_ref = None
        vmss_type_parts = base_name.split(".")
        vmss_type = vmss_type_parts[0]
        vmss_name = vmss_type_parts[1] if len(vmss_type_parts) > 1 else None

        if vmss_name:
            for file_path, resources in tfdata.get("all_resource", {}).items():
                if not isinstance(resources, list):
                    continue
                for res_block in resources:
                    if vmss_type in res_block and vmss_name in res_block.get(
                        vmss_type, {}
                    ):
                        vmss_data = res_block[vmss_type][vmss_name]
                        # Extract subnet from network_interface.ip_configuration.subnet_id
                        ni = vmss_data.get("network_interface", [{}])
                        if ni and isinstance(ni, list) and len(ni) > 0:
                            ipc = ni[0].get("ip_configuration", [{}])
                            if ipc and isinstance(ipc, list) and len(ipc) > 0:
                                subnet_id_raw = ipc[0].get("subnet_id", "")
                                # Extract subnet name from ${azurerm_subnet.main.id}
                                import re

                                match = re.search(
                                    r"\$\{([^.]+\.[^.]+)", str(subnet_id_raw)
                                )
                                if match:
                                    subnet_ref = match.group(1)
                        break
                if subnet_ref:
                    break

        if zones_attr and isinstance(zones_attr, list):

            # Only create zone containers if zones attribute exists and has values
            if zones_attr and isinstance(zones_attr, list) and subnet_ref:
                # Use the subnet reference from all_resource
                parent_subnet = subnet_ref if subnet_ref in subnets else None

                # Fallback: check which subnet contains any instance
                if not parent_subnet:
                    for subnet in subnets:
                        for instance in instances:
                            if instance in tfdata["graphdict"].get(subnet, []):
                                parent_subnet = subnet
                                break
                        if parent_subnet:
                            break

                if parent_subnet:
                    # Create zone container for each instance
                    for i, vmss_instance in enumerate(sorted(instances), start=1):
                        if i <= len(zones_attr):
                            zone_id = zones_attr[i - 1]
                            # Create zone node name
                            zone_node = f"tv_azurerm_zone.zone_{zone_id}"

                            # Create zone node in graphdict if it doesn't exist
                            if zone_node not in tfdata["graphdict"]:
                                tfdata["graphdict"][zone_node] = []
                                tfdata["meta_data"][zone_node] = {
                                    "zone_id": zone_id,
                                    "name": f"Availability Zone {zone_id}",
                                }

                            # Place VMSS instance inside zone
                            if vmss_instance not in tfdata["graphdict"][zone_node]:
                                tfdata["graphdict"][zone_node].append(vmss_instance)

                            # Remove VMSS instance from subnet direct children
                            if vmss_instance in tfdata["graphdict"].get(
                                parent_subnet, []
                            ):
                                tfdata["graphdict"][parent_subnet].remove(vmss_instance)

                            # Add zone to subnet if not already there
                            if zone_node not in tfdata["graphdict"].get(
                                parent_subnet, []
                            ):
                                tfdata["graphdict"][parent_subnet].append(zone_node)

                    # Remove the base VMSS node after creating numbered instances in zones
                    if base_name in tfdata["graphdict"]:
                        # Check if any resources connect TO the base node
                        base_connections = tfdata["graphdict"][base_name]

                        # Delete the base node from graphdict
                        del tfdata["graphdict"][base_name]

                        # Remove base node from any parent's children list
                        for node in tfdata["graphdict"]:
                            if base_name in tfdata["graphdict"][node]:
                                tfdata["graphdict"][node].remove(base_name)

                        # If the base node had connections, expand them to numbered instances
                        # For example, if base VMSS → Load Balancer, then vmss~1 → LB, vmss~2 → LB, vmss~3 → LB
                        if base_connections:
                            for instance in instances:
                                if instance not in tfdata["graphdict"]:
                                    tfdata["graphdict"][instance] = []
                                # Add base node's connections to each numbered instance
                                for conn in base_connections:
                                    if conn not in tfdata["graphdict"][instance]:
                                        tfdata["graphdict"][instance].append(conn)

    return tfdata


def place_vms_in_subnets(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Place VMs into subnets based on their NIC placements.

    This function runs AFTER create_multiple_resources so numbered VMs exist.
    It places VMs in the same subnet as their NICs for proper visual hierarchy.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with VMs placed in subnets
    """
    # Find all subnets
    subnets = [
        s
        for s in helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], "azurerm_subnet"
        )
        if "association" not in s
    ]

    # Find all VMs (numbered and unnumbered)
    vms = []
    for vm_type in [
        "azurerm_virtual_machine",
        "azurerm_linux_virtual_machine",
        "azurerm_windows_virtual_machine",
    ]:
        vms.extend(helpers.list_of_dictkeys_containing(tfdata["graphdict"], vm_type))

    # Find all NICs
    nics = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_network_interface"
    )
    nics = [n for n in nics if "association" not in n]

    # For each subnet, find VMs whose NICs are in that subnet
    for subnet in subnets:
        subnet_nics = tfdata["graphdict"].get(subnet, [])

        for vm in vms:
            # Check if this VM's NIC is in this subnet
            vm_metadata = tfdata["meta_data"].get(vm, {})
            vm_nic_ids = vm_metadata.get("network_interface_ids", [])

            # Match VM to NIC by checking if any NIC in subnet matches VM's NIC reference
            for nic in subnet_nics:
                if "azurerm_network_interface" in nic:
                    # Check if this NIC is referenced by the VM
                    nic_name = helpers.get_no_module_name(nic)
                    if nic in str(vm_nic_ids) or nic_name in str(vm_nic_ids):
                        # Place VM in subnet if not already there
                        if vm not in tfdata["graphdict"].get(subnet, []):
                            tfdata["graphdict"][subnet].append(vm)
                        break

    # Clean up base NIC nodes that were numbered - remove them from subnets
    # This handles the case where the base NIC was added to subnet before numbering occurred
    for subnet in subnets:
        subnet_children = tfdata["graphdict"].get(subnet, [])[:]
        for child in subnet_children:
            if (
                "azurerm_network_interface" in child
                and "~" not in child
                and "association" not in child
            ):
                # Check if this base NIC has numbered instances
                # Handle both formats: base~1 and base[0]~1 (from count)
                child_prefix = child.rsplit(".", 1)[0]  # Get resource type prefix
                child_name = (
                    child.rsplit(".", 1)[1] if "." in child else ""
                )  # Get resource name

                numbered_versions = [
                    k
                    for k in tfdata["graphdict"].keys()
                    if "~" in k and child_name in k and "azurerm_network_interface" in k
                ]
                # If numbered versions exist, remove the base node reference
                if numbered_versions:
                    tfdata["graphdict"][subnet].remove(child)

    return tfdata


def connect_lb_to_backend_vms(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Connect Load Balancers and Application Gateways to backend VMs.

    This function runs AFTER create_multiple_resources so numbered VMs and NICs exist.
    It creates direct connections from LB/AppGW to backend VMs, bypassing association
    resources which are implementation details.

    The connection path is: LB → Association → NIC → VM
    We want to show: LB → VM (direct logical connection)

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with LB connections to backend VMs
    """
    # Find all Load Balancers (consolidated from azurerm_lb and azurerm_lb_backend_address_pool)
    load_balancers = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_lb"
    )
    load_balancers = [
        lb
        for lb in load_balancers
        if "association" not in lb and "probe" not in lb and "rule" not in lb
    ]

    # Find all Application Gateways
    app_gateways = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_application_gateway"
    )

    # Find all association resources (numbered and base)
    associations = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"],
        "azurerm_network_interface_backend_address_pool_association",
    )

    # Find all VMs (numbered)
    vms = []
    for vm_type in [
        "azurerm_virtual_machine",
        "azurerm_linux_virtual_machine",
        "azurerm_windows_virtual_machine",
    ]:
        vms.extend(helpers.list_of_dictkeys_containing(tfdata["graphdict"], vm_type))

    # Find all NICs
    nics = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], "azurerm_network_interface"
    )
    nics = [nic for nic in nics if "association" not in nic]

    # Process each Load Balancer
    for lb in load_balancers:
        lb_connections = tfdata["graphdict"].get(lb, [])
        backend_vms = []

        # Check if LB connects to any association resources
        lb_associations = [conn for conn in lb_connections if "association" in conn]

        # Find backend VMs using the original Terraform metadata
        # Association resources link NICs to backend pools via count index
        # We need to find which VMs use which NICs

        # Get the backend pool metadata to understand the relationship
        # Look in all_resource for the association definition
        for file_path, resources in tfdata.get("all_resource", {}).items():
            if not isinstance(resources, list):
                continue
            for res_block in resources:
                # Check if this block defines an association resource
                if (
                    "azurerm_network_interface_backend_address_pool_association"
                    in res_block
                ):
                    assoc_configs = res_block.get(
                        "azurerm_network_interface_backend_address_pool_association", {}
                    )

                    # Iterate through each association resource definition (e.g., "main")
                    for assoc_name, assoc_config in assoc_configs.items():
                        nic_ref = assoc_config.get("network_interface_id", "")

                        # Extract the NIC resource name from the reference
                        # Format: ${azurerm_network_interface.backend[count.index].id}
                        nic_match = re.search(
                            r"azurerm_network_interface\.(\w+)", str(nic_ref)
                        )
                        if nic_match:
                            nic_base_name = nic_match.group(1)

                            # Find all NICs matching this base name
                            matching_nics = [
                                nic for nic in nics if f".{nic_base_name}" in nic
                            ]

                            # For each NIC, find the VM that uses it
                            # Match by index pattern - NIC backend[0]~1 matches VM backend[0]~1
                            for nic in matching_nics:
                                nic_pattern = nic.split("azurerm_network_interface.")[
                                    -1
                                ]  # Get "backend[0]~1"

                                for vm in vms:
                                    vm_pattern = (
                                        vm.split("azurerm_linux_virtual_machine.")[-1]
                                        if "linux" in vm
                                        else (
                                            vm.split(
                                                "azurerm_windows_virtual_machine."
                                            )[-1]
                                            if "windows" in vm
                                            else vm.split("azurerm_virtual_machine.")[
                                                -1
                                            ]
                                        )
                                    )

                                    # Match if they have the same index pattern (e.g., backend[0]~1)
                                    if nic_pattern == vm_pattern:
                                        if vm not in backend_vms:
                                            backend_vms.append(vm)
                                        break

        # Add direct connections from LB to VMs
        for vm in backend_vms:
            if vm not in tfdata["graphdict"][lb]:
                tfdata["graphdict"][lb].append(vm)

        # Remove association connections (they're implementation details)
        for assoc in lb_associations:
            if assoc in tfdata["graphdict"][lb]:
                tfdata["graphdict"][lb].remove(assoc)

    # Process Application Gateways - convert NIC connections to VM connections
    for appgw in app_gateways:
        appgw_connections = tfdata["graphdict"].get(appgw, [])
        backend_vms = []

        # Find NICs that AppGW connects to
        appgw_nics = [
            conn for conn in appgw_connections if "azurerm_network_interface" in conn
        ]

        # For each NIC, find the VM that uses it
        for nic in appgw_nics:
            nic_pattern = nic.split("azurerm_network_interface.")[
                -1
            ]  # Get "backend[0]~1"

            for vm in vms:
                vm_pattern = (
                    vm.split("azurerm_linux_virtual_machine.")[-1]
                    if "linux" in vm
                    else (
                        vm.split("azurerm_windows_virtual_machine.")[-1]
                        if "windows" in vm
                        else vm.split("azurerm_virtual_machine.")[-1]
                    )
                )

                # Match if they have the same index pattern (e.g., backend[0]~1)
                if nic_pattern == vm_pattern:
                    if vm not in backend_vms:
                        backend_vms.append(vm)
                    break

        # Add direct connections from AppGW to VMs
        for vm in backend_vms:
            if vm not in tfdata["graphdict"][appgw]:
                tfdata["graphdict"][appgw].append(vm)

        # Remove NIC connections (replace with VM connections)
        for nic in appgw_nics:
            if nic in tfdata["graphdict"][appgw]:
                tfdata["graphdict"][appgw].remove(nic)

    # Clean up association resources from the graph entirely
    for assoc in associations:
        tfdata["graphdict"].pop(assoc, None)
        tfdata["meta_data"].pop(assoc, None)

        # Remove from any parent connections
        for node in list(tfdata["graphdict"].keys()):
            if assoc in tfdata["graphdict"].get(node, []):
                tfdata["graphdict"][node].remove(assoc)

    return tfdata


def match_nsg_to_subnets(graphdict: Dict[str, List[str]]) -> Dict[str, List[str]]:
    """Match NSGs to subnets by suffix pattern.

    Args:
        graphdict: Resource graph dictionary

    Returns:
        Updated graphdict with NSG-subnet matches
    """
    result = dict(graphdict)
    suffix_pattern = r"~(\d+)$"

    # Find NSGs and subnets with numbered suffixes
    nsgs = [
        k
        for k in graphdict.keys()
        if "azurerm_network_security_group" in k and "association" not in k
    ]
    subnets = [
        k for k in graphdict.keys() if "azurerm_subnet" in k and "association" not in k
    ]

    for subnet in subnets:
        subnet_match = re.search(suffix_pattern, subnet)
        if subnet_match:
            subnet_suffix = subnet_match.group(1)

            # Find NSGs with matching suffix
            for nsg in nsgs:
                nsg_match = re.search(suffix_pattern, nsg)
                if nsg_match and nsg_match.group(1) == subnet_suffix:
                    if nsg not in result.get(subnet, []):
                        result[subnet].append(nsg)

    return result


def match_nic_to_vm(
    graphdict: Dict[str, List[str]], meta_data: Dict[str, Any]
) -> Dict[str, List[str]]:
    """Match NICs to VMs based on network_interface_ids reference.

    Args:
        graphdict: Resource graph dictionary
        meta_data: Resource metadata

    Returns:
        Updated graphdict with NIC-VM matches
    """
    result = dict(graphdict)

    nics = [
        k
        for k in graphdict.keys()
        if "azurerm_network_interface" in k and "association" not in k
    ]
    vms = [
        k
        for k in graphdict.keys()
        if any(
            vm_type in k
            for vm_type in [
                "azurerm_virtual_machine",
                "azurerm_linux_virtual_machine",
                "azurerm_windows_virtual_machine",
            ]
        )
    ]

    for vm in vms:
        if not meta_data.get(vm):
            continue

        nic_refs = str(meta_data[vm].get("network_interface_ids", ""))

        for nic in nics:
            nic_name = nic.split(".")[-1]
            if nic in nic_refs or nic_name in nic_refs:
                # Add VM to NIC's connections
                if vm not in result.get(nic, []):
                    result[nic].append(vm)

    return result


def remove_empty_groups(graphdict: Dict[str, List[str]]) -> Dict[str, List[str]]:
    """Remove empty group nodes from graph.

    Args:
        graphdict: Resource graph dictionary

    Returns:
        Updated graphdict with empty groups removed
    """
    result = dict(graphdict)

    # Find empty groups
    empty_groups = []
    for resource, connections in result.items():
        resource_type = helpers.get_no_module_name(resource).split(".")[0]
        if resource_type in GROUP_NODES and not connections:
            empty_groups.append(resource)

    # Remove empty groups
    for group in empty_groups:
        del result[group]

    return result



"""
Generic provides the possibility of load an image to be presented as a node.
"""

from resource_classes import Node


class _Generic(Node):
    provider = "generic"
    _icon_dir = "resource_images/generic"

    fontcolor = "#ffffff"



from . import _AWS


class _Storage(_AWS):
    _type = "storage"
    _icon_dir = "resource_images/aws/storage"


class Backup(_Storage):
    _icon = "backup.png"


class CloudendureDisasterRecovery(_Storage):
    _icon = "cloudendure-disaster-recovery.png"


class EFSInfrequentaccessPrimaryBg(_Storage):
    _icon = "efs-infrequentaccess-primary-bg.png"


class EFSStandardPrimaryBg(_Storage):
    _icon = "efs-standard-primary-bg.png"


class ElasticBlockStoreEBS(_Storage):
    _icon = "elastic-block-store-ebs.png"


class ElasticFileSystemEFS(_Storage):
    _icon = "elastic-file-system-efs.png"


class FsxForLustre(_Storage):
    _icon = "fsx-for-lustre.png"


class FsxForWindowsFileServer(_Storage):
    _icon = "fsx-for-windows-file-server.png"


class Fsx(_Storage):
    _icon = "fsx.png"


class S3Glacier(_Storage):
    _icon = "s3-glacier.png"


class SimpleStorageServiceS3(_Storage):
    _icon = "simple-storage-service-s3.png"


class SnowballEdge(_Storage):
    _icon = "snowball-edge.png"


class Snowball(_Storage):
    _icon = "snowball.png"


class Snowmobile(_Storage):
    _icon = "snowmobile.png"


class StorageGateway(_Storage):
    _icon = "storage-gateway.png"


class Storage(_Storage):
    _icon = "storage.png"


class S3Object(_Storage):
    _icon = "simple-storage-service-s3-object.png"


# Aliases

CDR = CloudendureDisasterRecovery
EBS = ElasticBlockStoreEBS
EFS = ElasticFileSystemEFS
FSx = Fsx
S3 = SimpleStorageServiceS3

# Terraform aliases
aws_backup_plan = Backup
aws_backup_vault = Backup
aws_backup_selection = Backup
aws_ebs_snapshot = ElasticBlockStoreEBS
aws_ebs_volume = ElasticBlockStoreEBS
aws_efs_access_point = ElasticFileSystemEFS
aws_efs_file_system = ElasticFileSystemEFS
aws_efs = ElasticFileSystemEFS
aws_fsx_lustre_file_system = FsxForLustre
aws_fsx_ontap_file_system = Fsx
aws_fsx_openzfs_file_system = Fsx
aws_fsx_windows_file_system = FsxForWindowsFileServer
aws_glacier_vault = S3Glacier
aws_s3_bucket = SimpleStorageServiceS3
aws_s3_bucket_object = SimpleStorageServiceS3
aws_s3_object = S3Object
aws_storagegateway_gateway = StorageGateway
aws_storagegateway_smb_file_share = StorageGateway
aws_storagegateway_nfs_file_share = StorageGateway



from . import _GCP


class _Iot(_GCP):
    _type = "iot"
    _icon_dir = "resources/gcp/iot"


class IotCore(_Iot):
    _icon = "iot-core.png"


# Aliases

# Terraform aliases
google_cloudiot_registry = IotCore
google_cloudiot_device = IotCore



from . import _AWS


class _Robotics(_AWS):
    _type = "robotics"
    _icon_dir = "resource_images/aws/robotics"


class RobomakerSimulator(_Robotics):
    _icon = "robomaker-simulator.png"


class Robomaker(_Robotics):
    _icon = "robomaker.png"


class Robotics(_Robotics):
    _icon = "robotics.png"


# Aliases

# Terraform aliases
aws_robomaker_robot_application = Robomaker
aws_robomaker_simulation_application = RobomakerSimulator
aws_robomaker_simulation_job = RobomakerSimulator



"""GCP resource-specific handlers for Terraform graph processing.

Handles special cases for GCP resources including projects, VPCs, subnets,
firewall rules, load balancers, and other GCP-specific relationships.
"""

from typing import Dict, List, Any
import modules.config.cloud_config_gcp as cloud_config
import modules.helpers as helpers
from ast import literal_eval
import re
import copy

REVERSE_ARROW_LIST = cloud_config.GCP_REVERSE_ARROW_LIST
IMPLIED_CONNECTIONS = cloud_config.GCP_IMPLIED_CONNECTIONS
GROUP_NODES = cloud_config.GCP_GROUP_NODES
CONSOLIDATED_NODES = cloud_config.GCP_CONSOLIDATED_NODES
NODE_VARIANTS = cloud_config.GCP_NODE_VARIANTS
SPECIAL_RESOURCES = cloud_config.GCP_SPECIAL_RESOURCES
SHARED_SERVICES = cloud_config.GCP_SHARED_SERVICES
DISCONNECT_SERVICES = cloud_config.GCP_DISCONNECT_LIST


def handle_special_cases(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle special resource cases and disconnections for GCP.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with special cases handled
    """
    # Remove connections to services specified in disconnect services
    for r in sorted(tfdata["graphdict"].keys()):
        for d in DISCONNECT_SERVICES:
            if d in r:
                tfdata["graphdict"][r] = []
    return tfdata


def gcp_handle_project(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP Project relationships.

    Projects are top-level containers in GCP. All GCP resources belong to a Project.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Project relationships configured
    """
    # TODO: Implement Project handling in Phase 4 (US2)
    # - Group all resources by their project attribute
    # - Create hierarchical structure: Project > VPC > Subnet > Resources
    return tfdata


def gcp_handle_vpc(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP VPC (Virtual Private Cloud) relationships.

    VPCs are global network containers in GCP.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with VPC relationships configured
    """
    # TODO: Implement VPC handling in Phase 4 (US2)
    # - Link VPCs to their Projects
    # - Group Subnets within VPCs
    # - VPCs are global in GCP (not regional)
    return tfdata


def gcp_handle_subnet(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP Subnet relationships.

    Subnets are regional network segments within VPCs in GCP.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Subnet relationships configured
    """
    # TODO: Implement Subnet handling in Phase 4 (US2)
    # - Link Subnets to their VPCs
    # - Group resources within Subnets
    # - Handle regional placement
    return tfdata


def gcp_handle_firewall(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP Firewall Rule relationships.

    Firewall rules are applied at the VPC level in GCP.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Firewall Rule relationships configured
    """
    # TODO: Implement Firewall handling in Phase 4 (US2)
    # - Associate Firewall Rules with VPCs
    # - Handle target tags and service accounts
    return tfdata


def gcp_handle_gke(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP GKE (Google Kubernetes Engine) cluster relationships.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with GKE relationships configured
    """
    # TODO: Implement GKE handling in Phase 4 (US2)
    # - Link GKE clusters to Subnets
    # - Handle node pools
    # - Link to Container Registry (GCR)
    return tfdata


def gcp_handle_instance_group(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP Compute Instance Group relationships.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Instance Group relationships configured
    """
    # TODO: Implement Instance Group handling in Phase 4 (US2)
    # - Link instance groups to zones
    # - Handle autoscaling
    return tfdata


def gcp_handle_backend_service(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Handle GCP Backend Service relationships (Load Balancing).

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with Backend Service relationships configured
    """
    # TODO: Implement Backend Service handling in Phase 4 (US2)
    # - Link to instance groups
    # - Link to health checks
    # - Link to forwarding rules
    return tfdata


def gcp_handle_sharedgroup(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Group shared GCP services into a shared services group.

    Shared services include KMS, Cloud Storage, Logging, Monitoring, GCR, etc.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with shared services grouped
    """
    # Find all shared services and group them
    for node in sorted(tfdata["graphdict"].keys()):
        substring_match = [s for s in SHARED_SERVICES if s in node]
        if substring_match:
            # Create shared services group if needed
            if not tfdata["graphdict"].get("google_group.shared_services"):
                tfdata["graphdict"]["google_group.shared_services"] = []
                tfdata["meta_data"]["google_group.shared_services"] = {}
            # Add node to shared services group
            if node not in tfdata["graphdict"]["google_group.shared_services"]:
                tfdata["graphdict"]["google_group.shared_services"].append(node)

    # Replace consolidated nodes with their consolidated names
    if tfdata["graphdict"].get("google_group.shared_services"):
        for service in sorted(
            list(tfdata["graphdict"]["google_group.shared_services"])
        ):
            if helpers.consolidated_node_check(service, tfdata):
                tfdata["graphdict"]["google_group.shared_services"] = list(
                    map(
                        lambda x: x.replace(
                            service, helpers.consolidated_node_check(service, tfdata)
                        ),
                        tfdata["graphdict"]["google_group.shared_services"],
                    )
                )
    return tfdata


def random_string_handler(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Remove random string resources from graph.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with random strings removed
    """
    randoms = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "random_string.")
    for r in list(randoms):
        del tfdata["graphdict"][r]
    return tfdata


def match_resources(tfdata: Dict[str, Any]) -> Dict[str, Any]:
    """Match GCP resources based on patterns and dependencies.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with resources matched
    """
    # TODO: Implement GCP-specific resource matching in Phase 4 (US2)
    # - Match firewall rules to VPCs
    # - Match instances to subnets
    # - Match load balancers to backend services
    return tfdata



"""File parser module for TerraVision.

This module handles parsing of Terraform files (.tf), variable files (.tfvars),
and annotation files (YAML). It discovers files from local directories or Git
repositories, parses HCL2 syntax, and extracts resources, modules, and variables.
"""

import fileinput
import os
import re
import tempfile
from pathlib import Path
from sys import exit
from typing import Dict, List, Tuple, Any, Optional

import click
import yaml
import hcl2

import modules.gitlibs as gitlibs

# Global module-level variables
annotations: Dict[str, Any] = dict()
start_dir: Path = Path.cwd()
temp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory(
    dir=tempfile.gettempdir()
)
abspath: str = os.path.abspath(__file__)
dname: str = os.path.dirname(abspath)
MODULE_DIR: str = str(Path(Path.home(), ".terravision", "module_cache"))

# Create module cache directory if it doesn't exist
if not os.path.exists(MODULE_DIR):
    os.makedirs(MODULE_DIR)

# Terraform sections to extract during parsing
EXTRACT: List[str] = [
    "module",
    "output",
    "variable",
    "locals",
    "resource",
    "data",
    "provider",
]


def find_tf_files(
    source: str,
    paths: Optional[List[str]] = None,
    mod: str = "main",
    recursive: bool = False,
) -> List[str]:
    """Discover Terraform files in local directory or Git repository.

    Searches for .tf files, .tfvars files, and annotation YAML files in the
    specified source location. Handles both local directories and Git URLs.

    Args:
        source: Local directory path or Git repository URL
        paths: Existing list of file paths to append to (default: empty list)
        mod: Module name for organizing cloned repositories (default: 'main')
        recursive: Whether to recursively search subdirectories (default: False)

    Returns:
        List of absolute paths to discovered Terraform files
    """
    global annotations

    if paths is None:
        paths = list()

    yaml_detected = False

    # Clone Git repository or use local directory
    if not os.path.isdir(source):
        source_location = gitlibs.clone_files(source, temp_dir.name, mod)
    else:
        source_location = source.strip()

    files = [f for f in os.listdir(source_location)]
    click.echo(f"  Added Source Location: {source}")

    # Scan for Terraform and annotation files
    for file in files:
        # Collect Terraform and variable files
        if (
            file.lower().endswith(".tf")
            or file.lower().endswith("auto.tfvars")
            or "terraform.tfvars" in file
        ):
            paths.append(os.path.join(source_location, file))

        # Load annotation YAML files if present
        if (
            file.lower().endswith("terravision.yml")
            or file.lower().endswith("architecture.yaml")
            and not yaml_detected
        ):
            full_filepath = Path(source_location).joinpath(file)
            with open(full_filepath, "r") as file:
                click.echo(f"  Detected architecture annotation file : {file.name} \n")
                yaml_detected = True
                annotations = yaml.safe_load(file)

    # Recursively search subdirectories if requested
    if recursive:
        for root, dir, files in os.walk(source_location):
            for d in dir:
                subdir = os.path.join(root, d)
                for file in os.listdir(subdir):
                    if file.lower().endswith(".tf") or file.lower().endswith(
                        "auto.tfvars"
                    ):
                        paths.append(os.path.join(subdir, file))

    # Validate that files were found
    if len(paths) == 0:
        click.echo(
            "ERROR: No Terraform .tf files found in current directory or your "
            "source location. Use --source parameter to specify directory or "
            "Github URL of source files"
        )
        exit()

    return paths


def handle_module(
    modules_list: List[Dict[str, Any]], tf_file_paths: List[str], filename: str
) -> Dict[str, Any]:
    """Process module declarations and map them to source locations.

    Creates a mapping between module names and their source directories,
    handling both local paths and remote module sources.

    Args:
        modules_list: List of module declaration dictionaries from HCL
        tf_file_paths: List of Terraform file paths
        filename: Source file containing the module declarations

    Returns:
        Dictionary with 'tf_file_paths' and 'module_source_dict' keys
    """
    temp_modules_dir = temp_dir.name
    module_source_dict: Dict[str, Dict[str, str]] = dict()

    # Map each module to its source location
    for i in range(len(modules_list)):
        module_stanza = modules_list[i]
        key = next(iter(module_stanza))
        module_source = module_stanza[key]["source"]

        # Handle remote vs local module sources
        if not module_source.startswith(".") and not module_source.startswith("\\"):
            # Remote module: create cache path
            localfolder = module_source.replace("/", "_")
            cache_path = str(
                os.path.join(temp_modules_dir, ";" + key + ";" + localfolder)
            )
            module_source_dict[key] = {
                "cache_path": str(cache_path),
                "source_file": filename,
            }
        else:
            # Local module: use source path directly
            module_source_dict[key] = {
                "cache_path": module_source,
                "source_file": filename,
            }

    return {"tf_file_paths": tf_file_paths, "module_source_dict": module_source_dict}


def iterative_parse(
    tf_file_paths: List[str],
    hcl_dict: Dict[str, Any],
    extract_sections: List[str],
    tfdata: Dict[str, Any],
    tf_mod_dir: str,
) -> Dict[str, Any]:
    """Parse Terraform files and extract resources, modules, and variables.

    Iteratively processes each Terraform file, parsing HCL2 syntax and extracting
    specified sections. Handles parsing errors and discovers nested modules.

    Args:
        tf_file_paths: List of Terraform file paths to parse
        hcl_dict: Dictionary to store parsed HCL content
        extract_sections: List of section names to extract (e.g., 'resource', 'module')
        tfdata: Main data dictionary to populate with parsed content
        tf_mod_dir: Directory containing Terraform modules

    Returns:
        Updated tfdata dictionary with parsed content
    """
    tfdata["module_source_dict"] = dict()

    # Parse each Terraform file
    for filename in tf_file_paths:
        filepath = Path(filename)
        fname = filepath.parent.name + "/" + filepath.name
        click.echo(f"  Parsing {filename}")

        # Attempt to parse HCL2 content
        with click.open_file(filename, "r", encoding="utf8") as f:
            try:
                hcl_dict[filename] = hcl2.load(f)
            except Exception as error:
                print("A Terraform HCL parsing error occurred:", filename, error)
                continue

            # Retry with character cleanup if initial parse failed
            if filename not in hcl_dict.keys():
                click.echo(
                    f"   WARNING: Unknown Error reading TF file {filename}. "
                    f"Attempting character cleanup fix.."
                )
                with tempfile.TemporaryDirectory(dir=temp_dir.name) as tempclean:
                    f_tmp = clean_file(filename, str(tempclean))
                    hcl_dict[filename] = hcl2.load(f_tmp)
                    if filename not in hcl_dict.keys():
                        click.echo(
                            f"   ERROR: Unknown Error reading TF file {filename}. "
                            f"Aborting!"
                        )
                        exit()

        # Extract specified sections from parsed HCL
        for section in extract_sections:
            if section in hcl_dict[filename]:
                section_name = "all_" + section
                if section_name not in tfdata.keys():
                    tfdata[section_name] = {}
                tfdata[section_name][filename] = hcl_dict[filename][section]
                click.echo(
                    click.style(
                        f"    Found {len(hcl_dict[filename][section])} {section} stanza(s)",
                        fg="green",
                    )
                )

                # Discover and process nested modules
                if section == "module":
                    for mod_dict in hcl_dict[filename]["module"]:
                        module_name = next(iter(mod_dict))
                        modpath = os.path.join(tf_mod_dir, module_name)
                        sourcemod = mod_dict[module_name]["source"]

                        # Handle relative module paths
                        if sourcemod.startswith("."):
                            curdir = os.getcwd()
                            os.chdir(os.path.dirname(filename))
                            modpath = os.path.abspath(sourcemod)
                            os.chdir(curdir)

                        # Fallback to source if module directory doesn't exist
                        if not os.path.isdir(modpath):
                            modpath = mod_dict[module_name]["source"]

                        # Recursively find files in module directory
                        source_files_list = find_tf_files(modpath, [], module_name)
                        existing_files = list(tf_file_paths)
                        tf_file_paths.extend(
                            x for x in source_files_list if x not in existing_files
                        )
                        tfdata["module_source_dict"][module_name] = str(modpath)

    # Handle duplicate module references
    oldpath: List[str] = []
    for module, modpath in tfdata["module_source_dict"].items():
        if modpath in oldpath:
            # Module called multiple times - duplicate resources
            duplicate = modpath
            for filepath, res_list in tfdata["all_resource"].items():
                if duplicate in filepath:
                    tfdata["all_resource"][filepath].append(
                        tfdata["all_resource"][filepath][0]
                    )
        else:
            oldpath.append(modpath)

    return tfdata


def read_tfsource(
    source_list: Tuple[str, ...],
    varfile_list: Tuple[str, ...],
    annotate: str,
    tfdata: Dict[str, Any],
) -> Dict[str, Any]:
    """Parse all Terraform files from source locations.

    Main entry point for parsing Terraform source files. Discovers and parses
    .tf files, loads variable files, and processes annotation files.

    Args:
        source_list: Tuple of source directory paths or Git URLs
        varfile_list: Tuple of variable file paths (.tfvars)
        annotate: Path to annotation YAML file (optional)
        tfdata: Dictionary to populate with parsed data

    Returns:
        Updated tfdata dictionary containing all parsed Terraform data
    """
    global annotations

    click.echo(click.style("\nParsing Terraform Source Files..", fg="white", bold=True))
    hcl_dict: Dict[str, Any] = dict()

    # Parse each source location
    for source in source_list:
        tf_file_paths = find_tf_files(source, [], False)
        tf_mod_dir = os.path.join(Path.home(), ".terraform", "modules")

        # Load custom annotation file if provided
        if annotate:
            with open(annotate, "r") as file:
                click.echo(f"  Will use architecture annotation file : {file.name} \n")
                annotations = yaml.safe_load(file)

        tfdata = iterative_parse(tf_file_paths, hcl_dict, EXTRACT, tfdata, tf_mod_dir)

    # Auto-detect and load .tfvars files
    for file in tf_file_paths:
        if "auto.tfvars" in file or "terraform.tfvars" in file:
            click.echo(f"  Will use auto variables from file : {file} \n")
            varfile_list = varfile_list + (file,)

    # Use all variable files if none specified
    if len(varfile_list) == 0 and tfdata.get("all_variable"):
        varfile_list = tuple(tfdata["all_variable"].keys())

    tfdata["varfile_list"] = list(varfile_list)
    tfdata["tempdir"] = temp_dir
    tfdata["annotations"] = annotations

    return tfdata


def clean_file(filename: str, tempdir: str):
    """Clean problematic characters from Terraform files.

    Attempts to fix HCL parsing errors by removing or escaping problematic
    characters and syntax that may cause parsing failures.

    Args:
        filename: Path to the Terraform file to clean
        tempdir: Temporary directory to write cleaned file

    Returns:
        File handle to the cleaned temporary file
    """
    filepath = str(Path(tempdir, "cleaning.tmp"))
    f_tmp = click.open_file(filepath, "w")

    with fileinput.FileInput(filename, inplace=False) as file:
        for line in file:
            # Skip comment lines
            if line.strip().startswith("#"):
                continue

            # Check for problematic characters
            if (
                '", "' in line
                or ":" in line
                or "*" in line
                or "?" in line
                or "[" in line
                or '("' in line
                or "==" in line
                or "]" in line
            ):
                # Clean AWS resource references with special characters
                if "aws_" in line and "resource" not in line:
                    array = line.split("=")
                    if len(array) > 1:
                        badstring = array[1]
                    else:
                        badstring = line
                    # Remove non-alphanumeric characters except dots and underscores
                    cleaned_string = re.sub("[^0-9a-zA-Z._]+", " ", badstring)
                    line = array[0] + ' = "' + cleaned_string + '"'
                else:
                    # Comment out problematic lines
                    line = f"# {line}" + "\r"

            f_tmp.write(line)

    # Reopen file for reading
    f_tmp = click.open_file(filepath, "r")
    return f_tmp



"""AWS-specific resource handler configurations.

Defines transformation pipelines for AWS resources in a config-driven manner.
Patterns support wildcards via substring matching.
"""

RESOURCE_HANDLER_CONFIGS = {
    "aws_eks_node_group": {
        "description": "Expand EKS node groups to numbered instances per subnet",
        "transformations": [
            {
                "operation": "expand_to_numbered_instances",
                "params": {
                    "resource_pattern": "aws_eks_node_group",
                    "subnet_key": "subnet_ids",
                    "skip_if_numbered": True,
                },
            },
        ],
    },
    "aws_eks_fargate_profile": {
        "description": "Expand Fargate profiles to numbered instances per subnet",
        "transformations": [
            {
                "operation": "expand_to_numbered_instances",
                "params": {
                    "resource_pattern": "aws_eks_fargate_profile",
                    "subnet_key": "subnet_ids",
                    "skip_if_numbered": True,
                },
            },
        ],
    },
    "aws_autoscaling_group": {
        "description": "Expand autoscaling groups to numbered instances per subnet",
        "transformations": [
            {
                "operation": "expand_to_numbered_instances",
                "params": {
                    "resource_pattern": "aws_autoscaling_group",
                    "subnet_key": "vpc_zone_identifier",
                    "skip_if_numbered": True,
                },
            },
        ],
    },
    "random_string": {
        "description": "Remove random string resources",
        "transformations": [
            {
                "operation": "delete_nodes",
                "params": {
                    "resource_pattern": "random_string.",
                    "remove_from_parents": True,
                },
            },
        ],
    },
    "aws_vpc_endpoint": {
        "description": "Move VPC endpoints into VPC parent and delete endpoint nodes",
        "transformations": [
            {
                "operation": "move_to_parent",
                "params": {
                    "resource_pattern": "aws_vpc_endpoint",
                    "from_parent_pattern": "aws_subnet",
                    "to_parent_pattern": "aws_vpc.",
                },
            },
            {
                "operation": "delete_nodes",
                "params": {
                    "resource_pattern": "aws_vpc_endpoint",
                    "remove_from_parents": False,
                },
            },
        ],
    },
    "aws_db_subnet_group": {
        "description": "Move DB subnet groups from subnets to VPC, redirect to security groups",
        "transformations": [
            {
                "operation": "move_to_vpc_parent",
                "params": {
                    "resource_pattern": "aws_db_subnet_group",
                },
            },
            {
                "operation": "redirect_to_security_group",
                "params": {
                    "resource_pattern": "aws_db_subnet_group",
                },
            },
        ],
    },
    "aws_": {
        "description": "Group shared AWS services into shared services group",
        "transformations": [
            {
                "operation": "group_shared_services",
                "params": {
                    "service_patterns": [
                        "aws_acm_certificate",
                        "aws_cloudwatch_log_group",
                        "aws_ecr_repository",
                        "aws_efs_file_system",
                        "aws_ssm_parameter",
                        "aws_kms_key",
                        "aws_eip",
                    ],
                    "group_name": "aws_group.shared_services",
                },
            },
        ],
    },
    "aws_cloudfront_distribution": {
        "description": "Connect CloudFront to load balancers and handle origins",
        "transformations": [
            {
                "operation": "link_via_shared_child",
                "params": {
                    "source_pattern": "aws_cloudfront",
                    "target_pattern": "aws_lb",
                    "remove_intermediate": False,
                },
            },
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_cloudfront",
                    "target_resource": "aws_acm_certificate.acm",
                    "metadata_key": "viewer_certificate",
                    "metadata_value_pattern": "acm_certificate_arn",
                },
            },
        ],
        "additional_handler_function": "handle_cf_origins",
    },
    "aws_subnet": {
        "description": "Create availability zone nodes and link to subnets",
        "handler_execution_order": "before",  # Run custom function FIRST to prepare metadata
        "additional_handler_function": "aws_prepare_subnet_az_metadata",
        "transformations": [
            # Insert AZ nodes between VPC and subnet (VPC→subnet becomes VPC→AZ→subnet)
            # This transformer handles both unlinking and relinking in one operation
            {
                "operation": "insert_intermediate_node",
                "params": {
                    "parent_pattern": "aws_vpc",
                    "child_pattern": "aws_subnet",
                    "intermediate_node_generator": "generate_az_node_name",
                    "create_if_missing": True,
                },
            },
        ],
    },
    "aws_lambda_function": {
        "description": "Move Lambda functions from subnets to VPC level to avoid duplication",
        "transformations": [
            {
                "operation": "move_to_vpc_parent",
                "params": {
                    "resource_pattern": "aws_lambda_function",
                },
            },
        ],
    },
    "aws_appautoscaling_target": {
        "description": "Handle autoscaling target relationships and counts",
        # Pure function handler - logic is too specific for generic transformers
        # Handles: 1) Count propagation from subnet to ASG/services
        #          2) Connection redirection (subnet→service becomes subnet→ASG)
        "additional_handler_function": "aws_handle_autoscaling",
    },
    "aws_efs_file_system": {
        "description": "Handle EFS mount targets and file system relationships",
        "transformations": [
            {
                "operation": "bidirectional_link",
                "params": {
                    "source_pattern": "aws_efs_mount_target",
                    "target_pattern": "aws_efs_file_system",
                    "cleanup_reverse": True,
                },
            },
        ],
        "additional_handler_function": "aws_handle_efs",
    },
    "aws_security_group": {
        "description": "Process security group relationships and reverse connections",
        "additional_handler_function": "aws_handle_sg",
    },
    "aws_lb": {
        "description": "Handle load balancer type variants and connections",
        "additional_handler_function": "aws_handle_lb",
    },
    "aws_ecs": {
        "description": "Handle ECS service configurations",
        "additional_handler_function": "aws_handle_ecs",
    },
    "aws_eks": {
        "description": "Handle EKS cluster and node group configurations",
        "additional_handler_function": "aws_handle_eks",
    },
    "helm_release": {
        "description": "Handle helm release resources",
        "additional_handler_function": "helm_release_handler",
    },
    "aws_elasticache_replication_group": {
        "description": "Pure Config-Driven: Expand replication groups to numbered instances per subnet",
        "transformations": [
            {
                "operation": "expand_to_numbered_instances",
                "params": {
                    "resource_pattern": "aws_elasticache_replication_group",
                    "subnet_key": "subnet_group_name",
                    "skip_if_numbered": True,
                    "inherit_connections": False,
                },
            },
            {
                "operation": "move_to_vpc_parent",
                "params": {
                    "resource_pattern": "aws_elasticache_subnet_group",
                },
            },
            {
                "operation": "redirect_to_security_group",
                "params": {
                    "resource_pattern": "aws_elasticache_subnet_group",
                },
            },
        ],
    },
    "aws_elasticache_cluster": {
        "description": "Pure Config-Driven: Expand ElastiCache clusters to numbered instances per subnet",
        "transformations": [
            {
                "operation": "expand_to_numbered_instances",
                "params": {
                    "resource_pattern": "aws_elasticache_cluster",
                    "subnet_key": "subnet_group_name",
                    "skip_if_numbered": True,
                    "inherit_connections": False,
                },
            },
        ],
    },
    "aws_wafv2_web_acl": {
        "description": "Hybrid: Link WAF Web ACLs to protected resources via association parsing",
        # Problem: WAF association doesn't create Terraform dependencies
        # Solution: Parse aws_wafv2_web_acl_association to create WAF → ALB/CloudFront connections
        "additional_handler_function": "aws_handle_waf_associations",
    },
    "aws_sagemaker_endpoint": {
        "description": "Config-Only: Delete endpoint_configuration (consolidation via AWS_CONSOLIDATED_NODES)",
        "transformations": [
            {
                "operation": "delete_nodes",
                "params": {
                    "resource_pattern": "aws_sagemaker_endpoint_configuration",
                    "remove_from_parents": True,
                },
            },
        ],
    },
    "aws_s3_bucket": {
        "description": "Pure Function: Group S3 buckets by region for cross-region replication scenarios",
        "additional_handler_function": "aws_handle_s3_cross_region_grouping",
    },
    "aws_s3_bucket_notification": {
        "description": "Config-Only: Link S3 notifications to targets + create transitive S3 bucket → target links",
        "transformations": [
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_s3_bucket_notification",
                    "target_resource": "aws_lambda_function",
                    "metadata_key": "lambda_function",
                    "metadata_value_pattern": "arn:aws:lambda",
                },
            },
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_s3_bucket_notification",
                    "target_resource": "aws_sns_topic",
                    "metadata_key": "topic",
                    "metadata_value_pattern": "arn:aws:sns",
                },
            },
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_s3_bucket_notification",
                    "target_resource": "aws_sqs_queue",
                    "metadata_key": "queue",
                    "metadata_value_pattern": "arn:aws:sqs",
                },
            },
            # Create transitive links: S3 bucket → Lambda (via notification intermediate)
            {
                "operation": "create_transitive_links",
                "params": {
                    "source_pattern": "aws_s3_bucket",
                    "intermediate_pattern": "aws_s3_bucket_notification",
                    "target_pattern": "aws_lambda_function",
                    "remove_intermediate": False,
                },
            },
            # Create transitive links: S3 bucket → SNS (via notification intermediate)
            {
                "operation": "create_transitive_links",
                "params": {
                    "source_pattern": "aws_s3_bucket",
                    "intermediate_pattern": "aws_s3_bucket_notification",
                    "target_pattern": "aws_sns_topic",
                    "remove_intermediate": False,
                },
            },
            # Create transitive links: S3 bucket → SQS (via notification intermediate)
            {
                "operation": "create_transitive_links",
                "params": {
                    "source_pattern": "aws_s3_bucket",
                    "intermediate_pattern": "aws_s3_bucket_notification",
                    "target_pattern": "aws_sqs_queue",
                    "remove_intermediate": False,
                },
            },
        ],
    },
    "aws_secretsmanager_secret": {
        "description": "Hybrid: Link Secrets Manager to rotation Lambda functions",
        "transformations": [
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_secretsmanager_secret_rotation",
                    "target_resource": "aws_lambda_function",
                    "metadata_key": "rotation_lambda_arn",
                    "metadata_value_pattern": "function:",
                },
            },
        ],
    },
    "aws_glue_catalog_table": {
        "description": "Hybrid: Link Glue Catalog tables to databases and S3 buckets",
        "transformations": [
            {
                "operation": "link_by_metadata_pattern",
                "params": {
                    "source_pattern": "aws_glue_catalog_table",
                    "target_resource": "aws_s3_bucket",
                    "metadata_key": "storage_descriptor",
                    "metadata_value_pattern": "s3://",
                },
            },
        ],
        "additional_handler_function": "aws_handle_glue_catalog",
    },
    "aws_appsync_graphql_api": {
        "description": "Config-Only: Consolidate AppSync resources + delete resolver nodes (consolidation via AWS_CONSOLIDATED_NODES)",
        "transformations": [
            {
                "operation": "delete_nodes",
                "params": {
                    "resource_pattern": "aws_appsync_resolver",
                    "remove_from_parents": True,
                },
            },
        ],
    },
    "aws_kinesis_firehose_delivery_stream": {
        "description": "Pure Function: Parse Firehose configuration for S3/Redshift/Elasticsearch destinations",
        # TEMPORARILY COMMENTED OUT FOR FUTURE IMPLEMENTATION
        # "additional_handler_function": "aws_handle_firehose",
    },
    "aws_lambda_event_source_mapping": {
        "description": "Pure Function: Create direct connections from event sources (SQS, Kinesis, DynamoDB) to Lambda functions",
        # Creates transitive links and removes intermediary mapping node
        # Handles multiple event source types (SQS, Kinesis, DynamoDB Streams)
        "additional_handler_function": "aws_handle_lambda_event_source_mapping",
    },
}



from . import _AWS


class _Mobile(_AWS):
    _type = "mobile"
    _icon_dir = "resource_images/aws/mobile"


class Amplify(_Mobile):
    _icon = "amplify.png"


class APIGatewayEndpoint(_Mobile):
    _icon = "api-gateway-endpoint.png"


class APIGatewayRed(_Mobile):
    _icon = "api-gateway.png"


class Appsync(_Mobile):
    _icon = "appsync.png"


class DeviceFarm(_Mobile):
    _icon = "device-farm.png"


class Pinpoint(_Mobile):
    _icon = "pinpoint.png"


# Aliases

# Terraform aliases
aws_amplify_app = Amplify
aws_amplify_branch = Amplify
aws_api_gateway_rest_api = APIGatewayRed
aws_apigatewayv2_api = APIGatewayRed
aws_appsync_graphql_api = Appsync
aws_appsync_function = Appsync
aws_devicefarm_project = DeviceFarm
aws_pinpoint_app = Pinpoint



from . import _GCP


class _Operations(_GCP):
    _type = "operations"
    _icon_dir = "resources/gcp/operations"


class Logging(_Operations):
    _icon = "logging.png"


class Monitoring(_Operations):
    _icon = "monitoring.png"


# Aliases

# Terraform aliases
google_logging_project_sink = Logging
google_logging_folder_sink = Logging
google_monitoring_alert_policy = Monitoring
google_monitoring_notification_channel = Monitoring
google_monitoring_uptime_check_config = Monitoring



from . import _Azure


class _Migration(_Azure):
    _type = "migration"
    _icon_dir = "resource_images/azure/migration"


class AzureDatabaseMigrationServices(_Migration):
    _icon = "azure-database-migration-services.png"


class DataBoxEdge(_Migration):
    _icon = "data-box-edge.png"


class DataBox(_Migration):
    _icon = "data-box.png"


class DatabaseMigrationServices(_Migration):
    _icon = "database-migration-services.png"


class MigrationProjects(_Migration):
    _icon = "migration-projects.png"


class RecoveryServicesVaults(_Migration):
    _icon = "recovery-services-vaults.png"


# Aliases

# Terraform aliases
azurerm_database_migration_service = DatabaseMigrationServices
azurerm_database_migration_project = MigrationProjects
azurerm_recovery_services_vault = RecoveryServicesVaults



from . import _Azure


class _Ml(_Azure):
    _type = "ml"
    _icon_dir = "resource_images/azure/ml"


class AzureOpenAI(_Ml):
    _icon = "azure-open-ai.png"


class AzureSpeechService(_Ml):
    _icon = "azure-speech-service.png"


class BatchAI(_Ml):
    _icon = "batch-ai.png"


class BotServices(_Ml):
    _icon = "bot-services.png"


class CognitiveServices(_Ml):
    _icon = "cognitive-services.png"


class GenomicsAccounts(_Ml):
    _icon = "genomics-accounts.png"


class MachineLearningServiceWorkspaces(_Ml):
    _icon = "machine-learning-service-workspaces.png"


class MachineLearningStudioWebServicePlans(_Ml):
    _icon = "machine-learning-studio-web-service-plans.png"


class MachineLearningStudioWebServices(_Ml):
    _icon = "machine-learning-studio-web-services.png"


class MachineLearningStudioWorkspaces(_Ml):
    _icon = "machine-learning-studio-workspaces.png"


# Aliases

# Terraform aliases
azurerm_cognitive_account = CognitiveServices
azurerm_bot_service_azure_bot = BotServices
azurerm_machine_learning_workspace = MachineLearningServiceWorkspaces



"""Generic resource transformation operations for AWS architecture diagrams.

Provides reusable building blocks for mutating tfdata['graphdict'] to create
professional AWS architecture diagrams. Each transformer is a pure function
that can be composed via configuration.
"""

from typing import Dict, List, Any, Callable, Optional
import copy
import re
import modules.helpers as helpers


def expand_to_numbered_instances(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    subnet_key: str = "subnet_ids",
    skip_if_numbered: bool = True,
    inherit_connections: bool = True,
) -> Dict[str, Any]:
    """Expand single resource into numbered instances (~1, ~2, ~3) per subnet.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources (e.g., "aws_eks_node_group")
        subnet_key: Metadata key containing subnet references
        skip_if_numbered: Skip resources already containing ~
        inherit_connections: If True, numbered instances inherit base resource connections.
                            If False, numbered instances start with empty connections.
                            Use False for visual-only resources (ElastiCache multi-AZ),
                            True for actual instances (EKS nodes, ASG instances).

    Returns:
        Updated tfdata with numbered instances
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )
    subnets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "aws_subnet")

    for resource in list(resources):
        if skip_if_numbered and "~" in resource:
            continue

        # Get subnet references from metadata
        subnet_ids = tfdata["meta_data"].get(resource, {}).get(subnet_key, [])
        if isinstance(subnet_ids, str):
            subnet_ids = [subnet_ids]

        # Find matching subnets
        matching_subnets = []
        for subnet in subnets:
            subnet_id = tfdata["meta_data"].get(subnet, {}).get("id", "")
            if any(subnet_id in str(sid) for sid in subnet_ids):
                matching_subnets.append(subnet)

        matching_subnets = sorted(matching_subnets)

        # Create numbered instances if multiple subnets
        if len(matching_subnets) > 1:
            for i, subnet in enumerate(matching_subnets, start=1):
                numbered = f"{resource}~{i}"
                # Inherit connections based on parameter
                if inherit_connections:
                    tfdata["graphdict"][numbered] = list(
                        tfdata["graphdict"].get(resource, [])
                    )
                else:
                    # Visual-only instances start with empty connections
                    tfdata["graphdict"][numbered] = []
                tfdata["meta_data"][numbered] = copy.deepcopy(
                    tfdata["meta_data"].get(resource, {})
                )

                # Add to subnet
                if numbered not in tfdata["graphdict"][subnet]:
                    tfdata["graphdict"][subnet].append(numbered)

                # Remove unnumbered from subnet
                if resource in tfdata["graphdict"][subnet]:
                    tfdata["graphdict"][subnet].remove(resource)

            # Delete original
            tfdata["graphdict"].pop(resource, None)
            tfdata["meta_data"].pop(resource, None)

        elif len(matching_subnets) == 1:
            # Single subnet - add if not present
            if resource not in tfdata["graphdict"][matching_subnets[0]]:
                tfdata["graphdict"][matching_subnets[0]].append(resource)

    # Cleanup: Ensure each subnet only contains numbered instances assigned to it
    # This prevents overlapping connections where subnet_a contains both resource~1 and resource~2
    subnets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "aws_subnet")
    for subnet_idx, subnet in enumerate(sorted(subnets), start=1):
        subnet_connections = tfdata["graphdict"].get(subnet, [])[:]

        for connection in subnet_connections:
            # Check if this is a numbered instance
            match = re.search(r"(.+)~(\d+)$", connection)
            if match:
                base_resource = match.group(1)
                instance_num = int(match.group(2))

                # If this numbered instance doesn't match the subnet's position, remove it
                if instance_num != subnet_idx:
                    tfdata["graphdict"][subnet].remove(connection)

    return tfdata


def apply_resource_variants(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    variant_map: Dict[str, str],
    metadata_key: str,
) -> Dict[str, Any]:
    """Apply variants to specific resource type and its connections.

    Replaces resource icons based on metadata and also updates connections.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources
        variant_map: Map of metadata values to new resource types
        metadata_key: Metadata key to check for variant

    Returns:
        Updated tfdata with variants applied
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )

    for resource in list(resources):
        if resource not in tfdata["meta_data"]:
            continue

        node_name = resource.split("~")[0] if "~" in resource else resource
        metadata_value = str(
            tfdata["meta_data"].get(node_name, {}).get(metadata_key, "")
        ).lower()

        # Find matching variant
        new_type = None
        for key, value in variant_map.items():
            if key.lower() in metadata_value:
                new_type = value
                break

        if new_type and resource in tfdata["graphdict"]:
            node_title = helpers.get_no_module_name(resource).split(".")[1]
            renamed = f"{new_type}.{node_title}"

            # Copy connections and metadata
            tfdata["graphdict"][renamed] = list(tfdata["graphdict"][resource])
            tfdata["meta_data"][renamed] = copy.deepcopy(
                tfdata["meta_data"].get(node_name, {})
            )

            # Update parent references
            for parent in helpers.list_of_parents(tfdata["graphdict"], resource):
                if resource in tfdata["graphdict"][parent]:
                    tfdata["graphdict"][parent].remove(resource)
                    if renamed not in tfdata["graphdict"][parent]:
                        tfdata["graphdict"][parent].append(renamed)

            # Delete original
            del tfdata["graphdict"][resource]

    return tfdata


def create_group_node(
    tfdata: Dict[str, Any],
    group_name: str,
    children: List[str],
    metadata: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Create a group/container node with children.

    Args:
        tfdata: Terraform data dictionary
        group_name: Name of group node to create
        children: List of child resource names
        metadata: Optional metadata for group

    Returns:
        Updated tfdata with group node
    """
    if group_name not in tfdata["graphdict"]:
        tfdata["graphdict"][group_name] = []
        tfdata["meta_data"][group_name] = metadata or {}

    for child in children:
        if child not in tfdata["graphdict"][group_name]:
            tfdata["graphdict"][group_name].append(child)

    return tfdata


def move_to_parent(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    from_parent_pattern: str,
    to_parent_pattern: str,
) -> Dict[str, Any]:
    """Move resources from one parent to another.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources to move
        from_parent_pattern: Pattern to match source parent
        to_parent_pattern: Pattern to match destination parent

    Returns:
        Updated tfdata with resources moved
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )
    from_parents = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], from_parent_pattern
    )
    to_parents = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], to_parent_pattern
    )

    if not to_parents:
        return tfdata

    to_parent = to_parents[0]

    for resource in resources:
        for from_parent in from_parents:
            if resource in tfdata["graphdict"].get(from_parent, []):
                tfdata["graphdict"][from_parent].remove(resource)
                if resource not in tfdata["graphdict"][to_parent]:
                    tfdata["graphdict"][to_parent].append(resource)

    return tfdata


def link_resources(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str,
    bidirectional: bool = False,
) -> Dict[str, Any]:
    """Create connections between resources.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_pattern: Pattern to match target resources
        bidirectional: Create connections in both directions

    Returns:
        Updated tfdata with new connections
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)
    targets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], target_pattern)

    for source in sources:
        for target in targets:
            if target not in tfdata["graphdict"].get(source, []):
                tfdata["graphdict"][source].append(target)

            if bidirectional and source not in tfdata["graphdict"].get(target, []):
                tfdata["graphdict"][target].append(source)

    return tfdata


def unlink_resources(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str,
) -> Dict[str, Any]:
    """Remove connections between resources.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_pattern: Pattern to match target resources

    Returns:
        Updated tfdata with connections removed
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)
    targets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], target_pattern)

    for source in sources:
        for target in targets:
            if target in tfdata["graphdict"].get(source, []):
                tfdata["graphdict"][source].remove(target)

    return tfdata


def delete_nodes(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    remove_from_parents: bool = True,
) -> Dict[str, Any]:
    """Delete nodes from graph.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources to delete
        remove_from_parents: Also remove from parent connections

    Returns:
        Updated tfdata with nodes deleted
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )

    for resource in list(resources):
        if remove_from_parents:
            parents = helpers.list_of_parents(tfdata["graphdict"], resource)
            for parent in parents:
                if resource in tfdata["graphdict"][parent]:
                    tfdata["graphdict"][parent].remove(resource)

        tfdata["graphdict"].pop(resource, None)
        tfdata["meta_data"].pop(resource, None)

    return tfdata


def match_by_suffix(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str,
) -> Dict[str, Any]:
    """Link resources with matching ~N suffixes.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_pattern: Pattern to match target resources

    Returns:
        Updated tfdata with suffix-matched connections
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)
    targets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], target_pattern)

    suffix_pattern = r"~(\d+)$"

    for source in sources:
        source_match = re.search(suffix_pattern, source)
        if not source_match:
            continue

        source_suffix = source_match.group(1)

        for target in targets:
            target_match = re.search(suffix_pattern, target)
            if target_match and target_match.group(1) == source_suffix:
                if target not in tfdata["graphdict"].get(source, []):
                    tfdata["graphdict"][source].append(target)

    return tfdata


def redirect_connections(
    tfdata: Dict[str, Any],
    from_resource_pattern: str,
    to_resource_pattern: str,
    parent_pattern: Optional[str] = None,
) -> Dict[str, Any]:
    """Redirect connections from one resource to another.

    Args:
        tfdata: Terraform data dictionary
        from_resource_pattern: Pattern to match resources to redirect from
        to_resource_pattern: Pattern to match resources to redirect to
        parent_pattern: Optional pattern to limit which parents are updated

    Returns:
        Updated tfdata with redirected connections
    """
    from_resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], from_resource_pattern
    )
    to_resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], to_resource_pattern
    )

    if not to_resources:
        return tfdata

    to_resource = to_resources[0]

    for from_resource in from_resources:
        parents = helpers.list_of_parents(tfdata["graphdict"], from_resource)

        if parent_pattern:
            parents = [p for p in parents if parent_pattern in p]

        for parent in parents:
            if from_resource in tfdata["graphdict"][parent]:
                tfdata["graphdict"][parent].remove(from_resource)
                if to_resource not in tfdata["graphdict"][parent]:
                    tfdata["graphdict"][parent].append(to_resource)

    return tfdata


def clone_with_suffix(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    count: int,
) -> Dict[str, Any]:
    """Clone resources with numbered suffixes.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources to clone
        count: Number of clones to create

    Returns:
        Updated tfdata with cloned resources
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )

    for resource in resources:
        if "~" in resource:
            continue

        for i in range(1, count + 1):
            numbered = f"{resource}~{i}"
            tfdata["graphdict"][numbered] = list(tfdata["graphdict"].get(resource, []))
            tfdata["meta_data"][numbered] = copy.deepcopy(
                tfdata["meta_data"].get(resource, {})
            )

        # Delete original
        tfdata["graphdict"].pop(resource, None)
        tfdata["meta_data"].pop(resource, None)

    return tfdata


def apply_all_variants(
    tfdata: Dict[str, Any],
) -> Dict[str, Any]:
    """Apply all resource variants based on NODE_VARIANTS config.

    Replaces handle_variants() function with config-driven approach.
    Handles both node renaming and connection renaming based on metadata.

    Args:
        tfdata: Terraform data dictionary

    Returns:
        Updated tfdata with all variants applied
    """
    from modules.provider_detector import get_primary_provider_or_default
    from modules.config_loader import load_config

    provider = get_primary_provider_or_default(tfdata)
    config = load_config(provider)

    # Get provider-specific constants
    provider_upper = provider.upper()
    NODE_VARIANTS = getattr(config, f"{provider_upper}_NODE_VARIANTS", {})
    SPECIAL_RESOURCES = getattr(config, f"{provider_upper}_SPECIAL_RESOURCES", {})
    provider_prefixes = config.PROVIDER_PREFIX

    # Process each node
    for node in list(tfdata["graphdict"].keys()):
        node_title = helpers.get_no_module_name(node).split(".")[1]
        node_name = (
            node.split("~")[0] if (node[-1].isdigit() and node[-2] == "~") else node
        )

        # Check if resource belongs to current provider
        resource_name = helpers.get_no_module_name(node_name)
        is_provider_resource = any(
            resource_name.startswith(prefix) for prefix in provider_prefixes
        )

        if not is_provider_resource:
            continue

        # Check for variant
        renamed_node = helpers.check_variant(node, tfdata["meta_data"].get(node_name))

        # Rename node if variant found and not already processed
        if (
            renamed_node
            and helpers.get_no_module_name(node).split(".")[0]
            not in SPECIAL_RESOURCES.keys()
            and node in tfdata["graphdict"]
        ):
            renamed_node = f"{renamed_node}.{node_title}"
            tfdata["graphdict"][renamed_node] = list(tfdata["graphdict"][node])
            tfdata["meta_data"][renamed_node] = copy.deepcopy(
                tfdata["meta_data"].get(node, {})
            )
            del tfdata["graphdict"][node]
            node = renamed_node

        # Process connections
        if node not in tfdata["graphdict"]:
            continue

        shared_group_name = f"{provider}_group.shared_services"

        for resource in list(tfdata["graphdict"][node]):
            connection_resource_name = (
                resource.split("~")[0]
                if ("~" in resource and resource[-1].isdigit() and resource[-2] == "~")
                else resource
            )

            # Check if connection belongs to current provider
            connection_name = helpers.get_no_module_name(connection_resource_name)
            is_provider_connection = any(
                connection_name.startswith(prefix) for prefix in provider_prefixes
            )

            if not is_provider_connection:
                continue

            variant_suffix = helpers.check_variant(
                resource, tfdata["meta_data"].get(connection_resource_name)
            )

            if (
                variant_suffix
                and helpers.get_no_module_name(resource).split(".")[0]
                not in SPECIAL_RESOURCES.keys()
                and not node.startswith(f"{provider}_group.shared")
                and (
                    shared_group_name not in tfdata["graphdict"]
                    or resource not in tfdata["graphdict"].get(shared_group_name, [])
                    or "~" in node
                )
                and resource.split(".")[0] != node.split(".")[0]
            ):
                variant_label = resource.split(".")[1]
                new_variant_name = f"{variant_suffix}.{variant_label}"

                tfdata["graphdict"][node].remove(resource)
                tfdata["graphdict"][node].append(new_variant_name)
                tfdata["meta_data"][new_variant_name] = copy.deepcopy(
                    tfdata["meta_data"].get(resource, {})
                )

    return tfdata


def move_to_vpc_parent(
    tfdata: Dict[str, Any],
    resource_pattern: str,
) -> Dict[str, Any]:
    """Move resources from subnets to VPC parent.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources to move

    Returns:
        Updated tfdata with resources moved to VPC
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )

    for resource in resources:
        parents = helpers.list_of_parents(tfdata["graphdict"], resource)
        subnets = [p for p in parents if "aws_subnet" in p]

        for subnet in subnets:
            if resource in tfdata["graphdict"][subnet]:
                tfdata["graphdict"][subnet].remove(resource)

                # Navigate to VPC through AZ
                az_parents = helpers.list_of_parents(tfdata["graphdict"], subnet)
                for az in az_parents:
                    if "aws_az" in az:
                        vpc_parents = helpers.list_of_parents(tfdata["graphdict"], az)
                        for vpc in vpc_parents:
                            if (
                                "aws_vpc" in vpc
                                and resource not in tfdata["graphdict"][vpc]
                            ):
                                tfdata["graphdict"][vpc].append(resource)

    return tfdata


def redirect_to_security_group(
    tfdata: Dict[str, Any],
    resource_pattern: str,
) -> Dict[str, Any]:
    """Redirect VPC connections to security group if resource has one.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match resources

    Returns:
        Updated tfdata with redirected connections
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )
    vpcs = helpers.list_of_dictkeys_containing(tfdata["graphdict"], "aws_vpc.")

    for resource in resources:
        for connection in tfdata["graphdict"].get(resource, []):
            # Check if connection has security group
            sg_parents = helpers.list_of_parents(tfdata["graphdict"], connection)
            security_groups = [p for p in sg_parents if "aws_security_group" in p]

            if security_groups:
                # Replace resource with security group in VPC
                for vpc in vpcs:
                    if resource in tfdata["graphdict"][vpc]:
                        tfdata["graphdict"][vpc].remove(resource)
                        for sg in security_groups:
                            if sg not in tfdata["graphdict"][vpc]:
                                tfdata["graphdict"][vpc].append(sg)

    return tfdata


def group_shared_services(
    tfdata: Dict[str, Any],
    service_patterns: List[str],
    group_name: str = "aws_group.shared_services",
) -> Dict[str, Any]:
    """Group shared services into a shared services group.

    Args:
        tfdata: Terraform data dictionary
        service_patterns: List of patterns to match shared services
        group_name: Name of the shared services group

    Returns:
        Updated tfdata with shared services grouped
    """
    # Create group if needed
    if group_name not in tfdata["graphdict"]:
        tfdata["graphdict"][group_name] = []
        tfdata["meta_data"][group_name] = {}

    # Find and add shared services (avoid duplicates)
    for node in sorted(tfdata["graphdict"].keys()):
        for pattern in service_patterns:
            if pattern in node and node not in tfdata["graphdict"][group_name]:
                tfdata["graphdict"][group_name].append(node)
                break

    # Replace consolidated nodes with their consolidated names
    updated_services = []
    for service in tfdata["graphdict"][group_name]:
        consolidated = helpers.consolidated_node_check(service, tfdata)
        if consolidated and "cluster" not in service:
            if consolidated not in updated_services:
                updated_services.append(consolidated)
        else:
            if service not in updated_services:
                updated_services.append(service)

    tfdata["graphdict"][group_name] = updated_services

    # Add default IAM service
    if "aws_iam_group.iam" not in tfdata["graphdict"][group_name]:
        tfdata["graphdict"][group_name].append("aws_iam_group.of_services")

    return tfdata


def link_via_shared_child(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str,
    remove_intermediate: bool = True,
) -> Dict[str, Any]:
    """Create direct links when source and target both connect to same intermediate.

    Detects pattern where both source and target connect to the same intermediate node,
    then creates a direct source → target connection.

    Example: CloudFront → Fargate, ALB → Fargate becomes CloudFront → ALB

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources (e.g., "aws_cloudfront")
        target_pattern: Pattern to match target resources (e.g., "aws_lb")
        remove_intermediate: Whether to remove intermediate connections and clean up

    Returns:
        Updated tfdata with direct links via shared children
    """
    from modules.config.cloud_config_aws import AWS_GROUP_NODES

    graphdict = tfdata["graphdict"]
    sources = sorted([s for s in graphdict.keys() if source_pattern in s])
    targets = sorted([s for s in graphdict.keys() if target_pattern in s])

    # For each node, check if it connects to both source and target
    for node in sorted(graphdict.keys()):
        connections = graphdict[node]

        for source in sources:
            if source in connections:
                for target in targets:
                    # Check if node also connects to target (node is parent of target)
                    if node in graphdict.get(target, []):
                        # Create direct source → target link
                        if target not in graphdict[source]:
                            graphdict[source].append(target)

                        if remove_intermediate:
                            # Remove source from intermediate node
                            graphdict[node].remove(source)

                            # Remove target from non-group parents
                            target_parents = helpers.list_of_parents(graphdict, target)
                            for parent in target_parents:
                                parent_type = helpers.get_no_module_name(parent).split(
                                    "."
                                )[0]
                                if parent_type not in AWS_GROUP_NODES:
                                    if target in graphdict[parent]:
                                        graphdict[parent].remove(target)

    return tfdata


def link_by_metadata_pattern(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_resource: str,
    metadata_key: str,
    metadata_value_pattern: str,
) -> Dict[str, Any]:
    """Create links when source metadata contains pattern.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_resource: Target resource to link to
        metadata_key: Metadata key to check
        metadata_value_pattern: Pattern to search for in metadata value

    Returns:
        Updated tfdata with metadata-based links
    """
    sources = sorted([s for s in tfdata["graphdict"].keys() if source_pattern in s])

    for source in sources:
        metadata = tfdata["meta_data"].get(source, {})
        metadata_value = str(metadata.get(metadata_key, ""))

        if metadata_value_pattern in metadata_value:
            if target_resource not in tfdata["graphdict"][source]:
                tfdata["graphdict"][source].append(target_resource)

    return tfdata


def create_transitive_links(
    tfdata: Dict[str, Any],
    source_pattern: str,
    intermediate_pattern: str,
    target_pattern: str,
    remove_intermediate: bool = True,
) -> Dict[str, Any]:
    """Create transitive links: a→b→c becomes a→c.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        intermediate_pattern: Pattern to match intermediate resources
        target_pattern: Pattern to match target resources
        remove_intermediate: Whether to remove intermediate node and connections

    Returns:
        Updated tfdata with transitive links created
    """
    graphdict = tfdata["graphdict"]
    sources = sorted([s for s in graphdict.keys() if source_pattern in s])
    intermediates = sorted([s for s in graphdict.keys() if intermediate_pattern in s])
    targets = sorted([s for s in graphdict.keys() if target_pattern in s])

    for source in sources:
        for intermediate in intermediates:
            if intermediate in graphdict.get(source, []):
                for target in targets:
                    if target in graphdict.get(intermediate, []):
                        # Create direct source → target link
                        if target not in graphdict[source]:
                            graphdict[source].append(target)

                        if remove_intermediate:
                            # Remove intermediate from source
                            graphdict[source].remove(intermediate)
                            # Remove intermediate node entirely
                            graphdict.pop(intermediate, None)
                            tfdata["meta_data"].pop(intermediate, None)

    return tfdata


def link_peers_via_intermediary(
    tfdata: Dict[str, Any],
    intermediary_pattern: str,
    source_pattern: str,
    target_pattern: str,
    remove_intermediary: bool = True,
) -> Dict[str, Any]:
    """Link peer resources that share an intermediary: intermediate→peer1 + intermediate→peer2 becomes peer1→peer2.

    Use case: Event source mappings, queue policies, and other configuration resources
    that logically connect two peers but shouldn't appear in diagrams.

    Pattern:
        Before: intermediate → source, intermediate → target
        After: source → target (intermediate removed)

    Args:
        tfdata: Terraform data dictionary
        intermediary_pattern: Pattern to match intermediary resources (e.g., "aws_lambda_event_source_mapping")
        source_pattern: Pattern to match source peer resources (e.g., "aws_sqs_queue")
        target_pattern: Pattern to match target peer resources (e.g., "aws_lambda_function")
        remove_intermediary: Whether to remove intermediary node (default: True)

    Returns:
        Updated tfdata with peer links created and intermediary optionally removed
    """
    graphdict = tfdata["graphdict"]
    intermediaries = sorted([s for s in graphdict.keys() if intermediary_pattern in s])

    for intermediary in list(intermediaries):
        intermediary_connections = graphdict.get(intermediary, [])

        if not intermediary_connections:
            continue

        # Find source and target peers in intermediary's connections
        sources = [c for c in intermediary_connections if source_pattern in c]
        targets = [c for c in intermediary_connections if target_pattern in c]

        # Only process if we have BOTH sources and targets
        if not (sources and targets):
            continue

        # Create direct links from each source to each target
        for source in sources:
            for target in targets:
                if source in graphdict:
                    if target not in graphdict[source]:
                        graphdict[source].append(target)
                else:
                    graphdict[source] = [target]

        # Remove intermediary node if requested (only if we created links)
        if remove_intermediary:
            # Remove the intermediary node itself
            graphdict.pop(intermediary, None)
            tfdata["meta_data"].pop(intermediary, None)

            # Clean up dangling references to the deleted intermediary from all other nodes
            for node in list(graphdict.keys()):
                if intermediary in graphdict[node]:
                    graphdict[node].remove(intermediary)

    return tfdata


def unlink_from_parents(
    tfdata: Dict[str, Any],
    resource_pattern: str,
    parent_filter: Optional[str] = None,
) -> Dict[str, Any]:
    """Remove child resources from all their parent connections.

    Args:
        tfdata: Terraform data dictionary
        resource_pattern: Pattern to match child resources
        parent_filter: Optional pattern to filter which parents to unlink from

    Returns:
        Updated tfdata with children unlinked from parents
    """
    resources = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], resource_pattern
    )

    for resource in resources:
        parents = helpers.list_of_parents(tfdata["graphdict"], resource)

        # Apply parent filter if specified
        if parent_filter:
            parents = [p for p in parents if parent_filter in p]

        for parent in parents:
            if resource in tfdata["graphdict"][parent]:
                tfdata["graphdict"][parent].remove(resource)

    return tfdata


def insert_intermediate_node(
    tfdata: Dict[str, Any],
    parent_pattern: str,
    child_pattern: str,
    intermediate_node_generator: Callable[[str, Dict[str, Any]], str],
    create_if_missing: bool = True,
) -> Dict[str, Any]:
    """Insert intermediate nodes between parents and children.

    Transforms: parent→child into parent→intermediate→child

    Args:
        tfdata: Terraform data dictionary
        parent_pattern: Pattern to match parent resources
        child_pattern: Pattern to match child resources
        intermediate_node_generator: Function that generates intermediate node name
                                     from child resource and its metadata
        create_if_missing: Create intermediate node if it doesn't exist

    Returns:
        Updated tfdata with intermediate nodes inserted
    """
    parents = helpers.list_of_dictkeys_containing(tfdata["graphdict"], parent_pattern)
    children = helpers.list_of_dictkeys_containing(tfdata["graphdict"], child_pattern)

    for child in children:
        child_parents = [p for p in parents if child in tfdata["graphdict"].get(p, [])]

        # Generate intermediate node name
        child_metadata = tfdata["meta_data"].get(child, {})
        intermediate = intermediate_node_generator(child, child_metadata)

        # Create intermediate node if needed
        if create_if_missing and intermediate not in tfdata["graphdict"]:
            tfdata["graphdict"][intermediate] = []
            tfdata["meta_data"][intermediate] = copy.deepcopy(child_metadata)

        # Rewire connections: parent→child becomes parent→intermediate→child
        for parent in child_parents:
            # Remove direct parent→child link
            if child in tfdata["graphdict"][parent]:
                tfdata["graphdict"][parent].remove(child)

            # Add parent→intermediate link
            if intermediate not in tfdata["graphdict"][parent]:
                tfdata["graphdict"][parent].append(intermediate)

        # Add intermediate→child link
        if child not in tfdata["graphdict"][intermediate]:
            tfdata["graphdict"][intermediate].append(child)

    return tfdata


def bidirectional_link(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str,
    cleanup_reverse: bool = False,
) -> Dict[str, Any]:
    """Create bidirectional links between resources.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_pattern: Pattern to match target resources
        cleanup_reverse: Remove target→source links after creating source→target

    Returns:
        Updated tfdata with bidirectional links
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)
    targets = helpers.list_of_dictkeys_containing(tfdata["graphdict"], target_pattern)

    for source in sources:
        for target in targets:
            # Create source→target link
            if target not in tfdata["graphdict"].get(source, []):
                tfdata["graphdict"][source].append(target)

            # Optionally clean up reverse link
            if cleanup_reverse and source in tfdata["graphdict"].get(target, []):
                tfdata["graphdict"][target].remove(source)

    return tfdata


def propagate_metadata(
    tfdata: Dict[str, Any],
    source_pattern: str,
    target_pattern: str = "",
    metadata_keys: Optional[List[str]] = None,
    direction: str = "forward",
    copy_from_connections: bool = False,
    propagate_to_children: bool = False,
) -> Dict[str, Any]:
    """Propagate metadata values from source to target resources.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        target_pattern: Pattern to match target resources (empty if propagate_to_children=True)
        metadata_keys: List of metadata keys to propagate (None = copy all keys)
        direction: "forward" (source→target), "reverse" (target→source), or "bidirectional"
        copy_from_connections: Copy metadata from resources connected to source
        propagate_to_children: Propagate to all children of source (ignores target_pattern)

    Returns:
        Updated tfdata with propagated metadata
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)

    # Determine targets based on mode
    if propagate_to_children:
        # Propagate to all children of sources
        targets = []
        for source in sources:
            children = tfdata["graphdict"].get(source, [])
            # Filter by target_pattern if provided
            if target_pattern:
                children = [c for c in children if target_pattern in c]
            targets.extend(children)
        targets = list(set(targets))  # Remove duplicates
    else:
        targets = helpers.list_of_dictkeys_containing(
            tfdata["graphdict"], target_pattern
        )

    for source in sources:
        source_metadata = tfdata["meta_data"].get(source, {})

        # When propagating to children, get children for this specific source
        if propagate_to_children:
            source_targets = tfdata["graphdict"].get(source, [])
            if target_pattern:
                source_targets = [t for t in source_targets if target_pattern in t]
        else:
            source_targets = targets

        for target in source_targets:
            if target not in tfdata["meta_data"]:
                tfdata["meta_data"][target] = {}

            # Determine which keys to copy
            keys_to_copy = (
                metadata_keys
                if metadata_keys is not None
                else list(source_metadata.keys())
            )

            # Propagate metadata keys
            for key in keys_to_copy:
                value = source_metadata.get(key)

                # Copy from connections if specified
                if copy_from_connections and not value:
                    for connection in tfdata["graphdict"].get(source, []):
                        conn_value = tfdata["meta_data"].get(connection, {}).get(key)
                        if conn_value:
                            value = conn_value
                            break

                if value is not None:
                    if direction in ["forward", "bidirectional"]:
                        tfdata["meta_data"][target][key] = copy.deepcopy(value)

                    if direction in ["reverse", "bidirectional"]:
                        tfdata["meta_data"][source][key] = copy.deepcopy(value)

    return tfdata


def replace_connection_targets(
    tfdata: Dict[str, Any],
    source_pattern: str,
    old_target_pattern: str,
    new_target_pattern: str,
) -> Dict[str, Any]:
    """Replace connection targets matching pattern with new targets.

    Args:
        tfdata: Terraform data dictionary
        source_pattern: Pattern to match source resources
        old_target_pattern: Pattern to match old connection targets
        new_target_pattern: Pattern to match new connection targets

    Returns:
        Updated tfdata with connection targets replaced
    """
    sources = helpers.list_of_dictkeys_containing(tfdata["graphdict"], source_pattern)
    old_targets = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], old_target_pattern
    )
    new_targets = helpers.list_of_dictkeys_containing(
        tfdata["graphdict"], new_target_pattern
    )

    if not new_targets:
        return tfdata

    new_target = new_targets[0]  # Use first matching new target

    for source in sources:
        for old_target in old_targets:
            if old_target in tfdata["graphdict"].get(source, []):
                tfdata["graphdict"][source].remove(old_target)
                if new_target not in tfdata["graphdict"][source]:
                    tfdata["graphdict"][source].append(new_target)

    return tfdata


def apply_transformation_pipeline(
    tfdata: Dict[str, Any],
    transformations: List[Dict[str, Any]],
) -> Dict[str, Any]:
    """Apply a sequence of transformations from configuration.

    Transformers are automatically discovered by name from this module's globals.
    The operation name in config must match the transformer function name exactly.

    Function parameters ending in '_function' or '_generator' are automatically
    resolved from string names to actual function references from handler modules.

    Args:
        tfdata: Terraform data dictionary
        transformations: List of transformation configs with 'operation' and 'params' keys
                        (params ending in _function or _generator are auto-resolved)

    Returns:
        Updated tfdata after all transformations

    Example:
        transformations = [
            {"operation": "expand_to_numbered_instances", "params": {...}},
            {"operation": "insert_intermediate_node", "params": {
                "intermediate_node_generator": "generate_az_node_name"  # Auto-resolved
            }},
        ]
    """
    for transform_config in transformations:
        operation = transform_config.get("operation")
        params = transform_config.get(
            "params", {}
        ).copy()  # Copy to avoid modifying original

        # Resolve function name strings to actual function references
        # Any parameter ending in _function or _generator is expected to be callable
        import modules.resource_handlers_aws as handlers_aws
        import modules.resource_handlers_gcp as handlers_gcp
        import modules.resource_handlers_azure as handlers_azure

        for param_name, param_value in params.items():
            if isinstance(param_value, str) and (
                param_name.endswith("_function") or param_name.endswith("_generator")
            ):
                # Try to find the function in handler modules
                if hasattr(handlers_aws, param_value):
                    params[param_name] = getattr(handlers_aws, param_value)
                elif hasattr(handlers_gcp, param_value):
                    params[param_name] = getattr(handlers_gcp, param_value)
                elif hasattr(handlers_azure, param_value):
                    params[param_name] = getattr(handlers_azure, param_value)
                else:
                    raise ValueError(
                        f"Could not resolve function '{param_value}' for parameter '{param_name}'. "
                        f"Make sure the function exists in one of the handler modules."
                    )

        # Dynamically look up transformer function by name
        # Operation name must match function name exactly
        transformer_func = globals().get(operation)

        if transformer_func and callable(transformer_func):
            tfdata = transformer_func(tfdata, **params)
        else:
            raise ValueError(
                f"Unknown transformer operation: '{operation}'. "
                f"Make sure the operation name matches a transformer function in resource_transformers.py"
            )

    return tfdata



from . import _GCP


class _Management(_GCP):
    _type = "management"
    _icon_dir = "resources/gcp/management"


class Billing(_Management):
    _icon = "billing.png"


class Project(_Management):
    _icon = "project.png"


class Quotas(_Management):
    _icon = "quotas.png"


class Support(_Management):
    _icon = "support.png"


# Aliases

# Terraform aliases
google_project = Project
google_billing_account = Billing



from . import _Azure


class _Security(_Azure):
    _type = "security"
    _icon_dir = "resource_images/azure/security"


class ApplicationSecurityGroups(_Security):
    _icon = "application-security-groups.png"


class AzureADAuthenticationMethods(_Security):
    _icon = "azure-ad-authentication-methods.png"


class AzureADIdentityProtection(_Security):
    _icon = "azure-ad-identity-protection.png"


class AzureADPrivlegedIdentityManagement(_Security):
    _icon = "azure-ad-privleged-identity-management.png"


class AzureADRiskySignins(_Security):
    _icon = "azure-ad-risky-signins.png"


class AzureADRiskyUsers(_Security):
    _icon = "azure-ad-risky-users.png"


class AzureInformationProtection(_Security):
    _icon = "azure-information-protection.png"


class AzureSentinel(_Security):
    _icon = "azure-sentinel.png"


class ConditionalAccess(_Security):
    _icon = "conditional-access.png"


class Defender(_Security):
    _icon = "defender.png"


class Detonation(_Security):
    _icon = "detonation.png"


class ExtendedSecurityUpdates(_Security):
    _icon = "extended-security-updates.png"


class Extendedsecurityupdates(_Security):
    _icon = "extendedsecurityupdates.png"


class IdentitySecureScore(_Security):
    _icon = "identity-secure-score.png"


class KeyVaults(_Security):
    _icon = "key-vaults.png"


class MicrosoftDefenderEasm(_Security):
    _icon = "microsoft-defender-easm.png"


class MicrosoftDefenderForCloud(_Security):
    _icon = "microsoft-defender-for-cloud.png"


class MicrosoftDefenderForIot(_Security):
    _icon = "microsoft-defender-for-iot.png"


class MultifactorAuthentication(_Security):
    _icon = "multifactor-authentication.png"


class SecurityCenter(_Security):
    _icon = "security-center.png"


class Sentinel(_Security):
    _icon = "sentinel.png"


class UserSettings(_Security):
    _icon = "user-settings.png"


# Aliases

# Terraform aliases
azurerm_application_security_group = ApplicationSecurityGroups
azurerm_sentinel_alert_rule = AzureSentinel
azurerm_sentinel_data_connector = AzureSentinel
azurerm_security_center_contact = SecurityCenter
azurerm_security_center_subscription_pricing = SecurityCenter
azurerm_key_vault = KeyVaults
azurerm_key_vault_key = KeyVaults
azurerm_key_vault_secret = KeyVaults



"""
Azure provides a set of services for Microsoft Azure provider.
"""

from resource_classes import Node


class _Azure(Node):
    _provider = "azure"
    _icon_dir = "resource_images/azure"

    def __init__(self, label="", **kwargs):
        # Azure service card styling - light grey rounded rectangles with internal padding
        card_style = {
            "shape": "box",
            "style": "rounded,filled",
            "fillcolor": "#F2F2F2",
            "color": "#E0E0E0",
            "penwidth": "1",
            "fontcolor": "#2C2C2C",
            "margin": "0.4",  # Internal margin/padding around the icon
        }
        # kwargs override card_style defaults
        merged_attrs = {**card_style, **kwargs}
        super().__init__(label, **merged_attrs)


class Azure(_Azure):
    _icon = "azure.png"



from . import _OnPrem


class _Security(_OnPrem):
    _type = "security"
    _icon_dir = "resource_images/onprem/security"


class Bitwarden(_Security):
    _icon = "bitwarden.png"


class Trivy(_Security):
    _icon = "trivy.png"


class Vault(_Security):
    _icon = "vault.png"


# Aliases



from . import _Azure


class _Storage(_Azure):
    _type = "storage"
    _icon_dir = "resource_images/azure/storage"


class ArchiveStorage(_Storage):
    _icon = "archive-storage.png"


class AzureDataboxGateway(_Storage):
    _icon = "azure-databox-gateway.png"


class AzureFileshares(_Storage):
    _icon = "azure-fileshares.png"


class AzureHcpCache(_Storage):
    _icon = "azure-hcp-cache.png"


class AzureNetappFiles(_Storage):
    _icon = "azure-netapp-files.png"


class AzureStackEdge(_Storage):
    _icon = "azure-stack-edge.png"


class Azurefxtedgefiler(_Storage):
    _icon = "azurefxtedgefiler.png"


class BlobStorage(_Storage):
    _icon = "blob-storage.png"


class DataBoxEdgeDataBoxGateway(_Storage):
    _icon = "data-box-edge-data-box-gateway.png"


class DataBox(_Storage):
    _icon = "data-box.png"


class DataLakeStorageGen1(_Storage):
    _icon = "data-lake-storage-gen1.png"


class DataLakeStorage(_Storage):
    _icon = "data-lake-storage.png"


class DataShareInvitations(_Storage):
    _icon = "data-share-invitations.png"


class DataShares(_Storage):
    _icon = "data-shares.png"


class GeneralStorage(_Storage):
    _icon = "general-storage.png"


class ImportExportJobs(_Storage):
    _icon = "import-export-jobs.png"


class NetappFiles(_Storage):
    _icon = "netapp-files.png"


class QueuesStorage(_Storage):
    _icon = "queues-storage.png"


class RecoveryServicesVaults(_Storage):
    _icon = "recovery-services-vaults.png"


class StorageAccountsClassic(_Storage):
    _icon = "storage-accounts-classic.png"


class StorageAccounts(_Storage):
    _icon = "storage-accounts.png"


class StorageExplorer(_Storage):
    _icon = "storage-explorer.png"


class StorageSyncServices(_Storage):
    _icon = "storage-sync-services.png"


class StorsimpleDataManagers(_Storage):
    _icon = "storsimple-data-managers.png"


class StorsimpleDeviceManagers(_Storage):
    _icon = "storsimple-device-managers.png"


class TableStorage(_Storage):
    _icon = "table-storage.png"


# Aliases

# Terraform aliases
azurerm_storage_account = StorageAccounts
azurerm_storage_blob = BlobStorage
azurerm_storage_container = BlobStorage
azurerm_storage_share = AzureFileshares
azurerm_storage_queue = QueuesStorage
azurerm_storage_table = TableStorage
azurerm_data_lake_store = DataLakeStorage
azurerm_netapp_account = NetappFiles
azurerm_netapp_volume = NetappFiles
azurerm_recovery_services_vault = RecoveryServicesVaults
azurerm_backup_vault = RecoveryServicesVaults
azurerm_storage_sync = StorageSyncServices



# AWS Cloud Configuration for TerraVision
# Provider: Amazon Web Services (aws provider)
# Architecture: VPC > Availability Zones > Subnets > Resources

from modules.config.resource_handler_configs_aws import RESOURCE_HANDLER_CONFIGS


# Provider metadata
PROVIDER_NAME = "AWS"
PROVIDER_PREFIX = ["aws_"]
ICON_LIBRARY = "aws"

# Any resource names with certain prefixes are consolidated into one node
AWS_CONSOLIDATED_NODES = [
    {
        "aws_route53": {
            "resource_name": "aws_route53_record.route_53",
            "import_location": "resource_classes.aws.network",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_cloudwatch_log": {
            "resource_name": "aws_cloudwatch_log_group.cloudwatch",
            "import_location": "resource_classes.aws.management",
            "vpc": False,
        }
    },
    {
        "aws_cloudwatch_event": {
            "resource_name": "aws_cloudwatch_event_rule.eventbridge",
            "import_location": "resource_classes.aws.integration",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_sns_topic": {
            "resource_name": "aws_sns_topic.sns",
            "import_location": "resource_classes.aws.integration",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_api_gateway": {
            "resource_name": "aws_api_gateway_integration.gateway",
            "import_location": "resource_classes.aws.network",
            "vpc": False,
        }
    },
    {
        "aws_acm": {
            "resource_name": "aws_acm_certificate.acm",
            "import_location": "resource_classes.aws.security",
            "vpc": False,
        }
    },
    {
        "aws_ssm_parameter": {
            "resource_name": "aws_ssm_parameter.ssmparam",
            "import_location": "resource_classes.aws.management",
            "vpc": False,
        }
    },
    {
        "aws_dx": {
            "resource_name": "aws_dx_connection.directconnect",
            "import_location": "resource_classes.aws.network",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_lb": {
            "resource_name": "aws_lb.elb",
            "import_location": "resource_classes.aws.network",
            "vpc": True,
        }
    },
    {
        "aws_ecs": {
            "resource_name": "aws_ecs_service.ecs",
            "import_location": "resource_classes.aws.compute",
            "vpc": True,
        }
    },
    {
        "aws_internet_gateway": {
            "resource_name": "aws_internet_gateway.igw",
            "import_location": "resource_classes.aws.network",
            "vpc": True,
        }
    },
    {
        "aws_efs_file_system": {
            "resource_name": "aws_efs_file_system.efs",
            "import_location": "resource_classes.aws.storage",
            "vpc": False,
        }
    },
    {
        "aws_kms": {
            "resource_name": "aws_kms_key.kms",
            "import_location": "resource_classes.aws.kms",
            "vpc": False,
        }
    },
    {
        "aws_eip": {
            "resource_name": "aws_eip.elastic_ip",
            "import_location": "resource_classes.eip.eip",
            "vpc": False,
        }
    },
    {
        "aws_autoscaling_policy": {
            "resource_name": "aws_autoscaling_policy.autoscaling_policy",
            "import_location": "resource_classes.aws.compute",
            "vpc": True,
        }
    },
    {
        "aws_sagemaker_endpoint": {
            "resource_name": "aws_sagemaker_endpoint.endpoint",
            "import_location": "resource_classes.aws.ml",
            "vpc": False,
        }
    },
    {
        "aws_appsync_graphql_api": {
            "resource_name": "aws_appsync_graphql_api.graphql_api",
            "import_location": "resource_classes.aws.network",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_cognito": {
            "resource_name": "aws_cognito_user_pool.cognito",
            "import_location": "resource_classes.aws.security",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_wafv2": {
            "resource_name": "aws_wafv2_web_acl.waf",
            "import_location": "resource_classes.aws.security",
            "vpc": False,
            "edge_service": True,
        }
    },
    {
        "aws_waf": {
            "resource_name": "aws_waf_web_acl.waf",
            "import_location": "resource_classes.aws.security",
            "vpc": False,
            "edge_service": True,
        }
    },
]

# List of Group type nodes and order to draw them in
AWS_GROUP_NODES = [
    "aws_vpc",
    "aws_az",
    "aws_group",
    "aws_account",
    "aws_appautoscaling_target",
    "aws_autoscaling_group",
    "aws_subnet",
    "aws_security_group",
    "tv_aws_onprem",
    "tv_aws_region",
]

# Nodes to be drawn first inside the AWS Cloud but outside any subnets or VPCs
AWS_EDGE_NODES = [
    "aws_route53",
    "aws_cloudfront_distribution",
    "aws_internet_gateway",
    "aws_api_gateway",
    "aws_apigateway",
    "aws_cloudwatch_event",
    "aws_sns_topic",
    "aws_cognito",
    "aws_wafv2",
    "aws_waf",
    "aws_appsync",
]

# Nodes outside Cloud boundary
AWS_OUTER_NODES = [
    "tv_aws_users",
    "tv_aws_internet",
    "tv_aws_device",
    "tv_aws_onprem",
    "tv_aws_mobile_client",
]

# Order to draw nodes - leave empty string list till last to denote everything else
AWS_DRAW_ORDER = [
    AWS_OUTER_NODES,
    AWS_EDGE_NODES,
    AWS_GROUP_NODES,
    AWS_CONSOLIDATED_NODES,
    [""],
]

# List of prefixes where additional nodes should be created automatically
AWS_AUTO_ANNOTATIONS = [
    {"aws_route53": {"link": ["tv_aws_users.users"], "arrow": "reverse"}},
    {
        "aws_cloudfront_distribution": {
            "link": ["tv_aws_users.users"],
            "arrow": "reverse",
        }
    },
    {
        "aws_dx": {
            "link": [
                "tv_aws_onprem.corporate_datacenter",
                "tv_aws_cgw.customer_gateway",
            ],
            "arrow": "forward",
        }
    },
    {
        "aws_internet_gateway": {
            "link": ["tv_aws_internet.internet"],
            "delete": ["aws_nat_gateway."],
            "arrow": "forward",
        }
    },
    {"aws_eks_cluster": {"link": ["aws_eks_service.eks"], "arrow": "reverse"}},
    {"aws_nat_gateway": {"link": ["aws_internet_gateway.*"], "arrow": "forward"}},
    {"aws_ecs_service": {"link": ["aws_ecr_repository.ecr"], "arrow": "forward"}},
    {"aws_eks_cluster": {"link": ["aws_ecr_repository.ecr"], "arrow": "forward"}},
    {"aws_api_gateway": {"link": ["tv_aws_mobile_client.mobile"], "arrow": "reverse"}},
    {"aws_ecs_": {"link": ["aws_ecs_cluster.ecs"], "arrow": "forward"}},
    {
        "aws_lambda": {
            "link": ["aws_cloudwatch_log_group.cloudwatch"],
            "arrow": "forward",
        }
    },
]

# Variant icons for the same service - matches keyword in meta data and changes resource type
AWS_NODE_VARIANTS = {
    "aws_ecs_service": {"FARGATE": "aws_fargate", "EC2": "aws_ec2ecs"},
    "aws_eks_cluster": {"compute_config": "aws_eks_cluster_auto"},
    "aws_lb": {"application": "aws_alb", "network": "aws_nlb"},
    "aws_rds": {
        "aurora": "aws_rds_aurora",
        "mysql": "aws_rds_mysql",
        "postgres": "aws_rds_postgres",
    },
}

# Automatically reverse arrow direction for these resources when discovered through source
AWS_REVERSE_ARROW_LIST = [
    "aws_route53",
    "aws_cloudfront",
    "aws_cloudwatch_event",  # EventBridge emits events TO Lambda (reverse direction)
    "aws_sfn_state_machine",  # Step Functions orchestrates services (reverse direction)
    "aws_vpc.",
    "aws_subnet.",
    "aws_appautoscaling_target",
    "aws_iam_role.",
    "aws_rds_aurora",
]

# Force certain resources to be a destination connection only - original TF node relationships only
AWS_FORCED_DEST = ["aws_rds", "aws_instance", "aws_elasticache"]

# Force certain resources to be a origin connection only - original TF node relationships only
AWS_FORCED_ORIGIN = [
    "aws_route53",
    "aws_cloudfront_distribution",
    "aws_cloudwatch_event",  # EventBridge emits events (source only, not destination)
    "aws_sns_topic",  # SNS emits messages to subscribers (source only, not destination)
    "aws_sfn_state_machine",  # Step Functions orchestrates services (source only, not destination)
    "aws_s3_bucket_notification",  # S3 notifications trigger services (source only, not destination)
    "aws_wafv2_web_acl",  # WAF protects resources (source only, not destination)
    "aws_waf_web_acl",  # WAF Classic protects resources (source only, not destination)
]


AWS_IMPLIED_CONNECTIONS = {
    "certificate_arn": "aws_acm_certificate",
    "container_definitions": "aws_ecr_repository",
}

# Generate AWS_SPECIAL_RESOURCES from RESOURCE_HANDLER_CONFIGS for backward compatibility with older functions
# Include any resource pattern that has transformations or additional handlers
AWS_SPECIAL_RESOURCES = {
    pattern: config.get("additional_handler_function", f"config_handler_{pattern}")
    for pattern, config in RESOURCE_HANDLER_CONFIGS.items()
}

AWS_SHARED_SERVICES = [
    "aws_acm_certificate",
    "aws_cloudwatch_log_group",
    "aws_ecr_repository",
    "aws_efs_file_system",
    "aws_ssm_parameter",
    "aws_kms_key",
    "aws_eip",
]

AWS_ALWAYS_DRAW_LINE = [
    "aws_lb",
    "aws_iam_role",
    "aws_volume_attachment",
    "aws_alb",
    "aws_nlb",
    "aws_efs_mount_target",
    "aws_ecs_service",
    "aws_rds_aurora",
    "aws_rds_mysql",
    "aws_rds_postgres",
]
# Resources that should never have lines drawn between them
AWS_NEVER_DRAW_LINE = []

# Resources that should be disconnected
AWS_DISCONNECT_LIST = []

# Resources that should be hidden from the diagram by default
AWS_HIDE_NODES = ["aws_security_group_rule"]

# Resources that should skip automatic expansion in handle_singular_references
# These resources are manually matched to subnets by suffix in their handlers
AWS_SKIP_SINGULAR_EXPANSION = [
    "aws_eks_fargate_profile",
    "aws_eks_node_group",
]

AWS_ACRONYMS_LIST = [
    "acm",
    "acm",
    "alb",
    "api",
    "db",
    "dx",
    "ebs",
    "ec2",
    "ecr",
    "ecs",
    "efs",
    "eip",
    "eks",
    "elb",
    "etl",
    "igw",
    "iam",
    "ip",
    "kms",
    "lb",
    "nat",
    "nlb",
    "rds",
    "s3",
    "sns",
    "sqs",
    "vpc",
]

AWS_NAME_REPLACEMENTS = {
    "az": "Availability Zone",
    "alb": "App Load Balancer",
    "appautoscaling_target": "Auto Scaling",
    "route_table_association": "Route Table",
    "ecs_service_fargate": "Fargate",
    "eip": "Elastic IP",
    "instance": "EC2",
    "lambda_function": "Lambda",
    "iam_role": "Role",
    "dx": "Direct Connect",
    "cloudfront_distribution": "Cloudfront",
    "iam_policy": "policy",
    "this": "",
}


AWS_REFINEMENT_PROMPT = """
You are an expert AWS Solutions Architect. I have a JSON representation of an AWS architecture diagram generated from Terraform code. The diagram may have incorrect resource groupings, missing connections, or layout issues.
INPUT JSON FORMAT: Each key is a Terraform resource ID, and its value is a list of resource IDs it connects to.
SPECIAL CONVENTIONS:
- Resources starting with "tv_" are visual helper nodes (e.g., "tv_aws_internet.internet" represents the public internet)
- Groups like VPCs, Subnets, Autoscaling, Security Groups etc are always parents and a connection to a resource means the resource is inside that group. However, if a resource inside a group is not also a group type, it should have no connections back to the group nodes
- "aws_az.availability_zone_*" groups represent availability zone boundaries
- Security Groups should always be a group containing EC2 instances or other resources within a subnet
- Resources ending with ~1, ~2, ~3 (instance number) indicate they're either multiple instances of the same resource, or one resource duplicated for clarity when a resource is deployed in multiple availability zones and subnets
- aws_group.shared_services is used to group common shared resources which are accessed by multiple resources such as CloudWatch, ECR, KMS and do not have any incoming or outgoing connections 
Please refine this diagram following AWS conventions and industry best practices:
1. Fix resource groupings (VPCs, subnets, availability zones, security groups) where necessary
2. Add missing logical connections between resources
3. Remove incorrect connections
4. Ensure proper hierarchy (Region > VPC > AZ > Subnet > Resources)
5. Group related resources (e.g., ALB with target groups, Autoscaling resources in an AutoScaling Group, EKS nodes within a subnet)
6. Group type resources like aws_group, aws_appautoscaling_targ, aws_subnet, aws_security_group and tv_aws_oprem are always the parent and other non group resources should be children without connections back to the group
7. Add implied connections (e.g., Lambda in VPC needs ENI connection, ECS or EKS will connect to an ECR repository node)
8. Autoscaling targets should appear as group type nodes within their relevant subnet with the same instance number
9. NAT Gateways in a public subnet should be duplicated with incremental instance numbers in all public subnets to show multi-AZ deployment
10. EKS Cluster worker nodes should be inside subnet groups

"""


AWS_DOCUMENTATION_PROMPT = """\
You are an AWS architect that needs to summarise this JSON of Terraform AWS resources and their associations concisely in paragraph form using as few bullet points as possible. Follow these instructions:
1. If you see ~1, ~2, ~3 etc at the end of the resource name it means multiple instances of the same resource are created. Include how many of each resource type are created in the summary. 
2. Use only AWS resource names in the text which can be inferred from terraform resource type names. e.g. instead of aws_ec2_instance.XXX just say an EC2 Instance named XXX
3. Mention which resources are associated with each respective subnet and availability zone if any. 
4. Provide an overall summary of the architecture and what the system does

"""

# Configuration patterns for multi-instance resource detection
# Each pattern defines:
# - resource_types: List of Terraform resource types to check
# - trigger_attributes: Attributes that trigger expansion (e.g., "subnets", "zones")
# - also_expand_attributes: Attributes containing related resources to also expand
# - resource_pattern: Regex pattern to extract resource references from attribute values
AWS_MULTI_INSTANCE_PATTERNS = [
    {
        "resource_types": ["aws_lb", "aws_alb", "aws_nlb"],
        "trigger_attributes": ["subnets"],
        "also_expand_attributes": ["security_groups"],
        "resource_pattern": r"\$\{(aws_\w+\.\w+)",
        "description": "ALB/NLB spanning multiple subnets",
    },
    {
        "resource_types": ["aws_ecs_service"],
        "trigger_attributes": ["subnets"],
        "also_expand_attributes": ["security_groups"],
        "resource_pattern": r"\$\{(aws_\w+\.\w+)",
        "description": "ECS service spanning multiple subnets",
    },
    # Add more AWS patterns as needed
]

# Replace with your OLLAMA server IP and port number
OLLAMA_HOST = "http://localhost:11434"

# Replace with your actual API Gateway endpoint from: terraform output api_endpoint
BEDROCK_API_ENDPOINT = (
    "https://yirz70b5mc.execute-api.us-east-1.amazonaws.com/prod/chat"
)
